"""This is ported py-version of the pru-code.

Goals:

- stay close to original code-base
- offer a comparison for the tests
- step 1 to a virtualization of emulation

NOTE: DO NOT OPTIMIZE -> stay close to original code-base

Compromises:

- bitshifted values (i.e. _n28) are converted to float without shift

"""

import math

from shepherd_core.data_models import CalibrationEmulator
from shepherd_core.data_models.content.virtual_source_config import LUT_SIZE
from shepherd_core.data_models.content.virtual_source_config import ConverterPRUConfig
from shepherd_core.data_models.content.virtual_storage_config import StoragePRUConfig

from .virtual_storage_model import VirtualStorageModelPRU


class PruCalibration:
    """part of calibration.h."""

    # negative residue compensation - compensate for noise around 0
    # -> current uint-design cuts away negative part and leads to biased mean()
    NOISE_ESTIMATE_nA: int = 2000
    RESIDUE_SIZE_FACTOR: int = 30
    RESIDUE_MAX_nA: int = NOISE_ESTIMATE_nA * RESIDUE_SIZE_FACTOR
    negative_residue_nA = 0

    def __init__(self, cal_emu: CalibrationEmulator | None = None) -> None:
        self.cal = cal_emu if cal_emu else CalibrationEmulator()

    def conv_adc_raw_to_nA(self, current_raw: int) -> float:
        I_nA = self.cal.adc_C_A.raw_to_si(current_raw) * (10**9)
        if self.cal.adc_C_A.offset < 0:
            if I_nA > self.negative_residue_nA:
                I_nA -= self.negative_residue_nA
                self.negative_residue_nA = 0
            else:
                self.negative_residue_nA = self.negative_residue_nA - I_nA
                self.negative_residue_nA = min(self.negative_residue_nA, self.RESIDUE_MAX_nA)
                I_nA = 0
        return I_nA

    @staticmethod
    def conv_adc_raw_to_uV(voltage_raw: int) -> None:
        msg = f"This Fn should not been used (val={voltage_raw})"
        raise RuntimeError(msg)

    def conv_uV_to_dac_raw(self, voltage_uV: float) -> int:
        dac_raw = self.cal.dac_V_A.si_to_raw(float(voltage_uV) / (10**6))
        return min(dac_raw, (2**16) - 1)


class VirtualConverterModel:
    """Ported python version of the pru vCnv."""

    def __init__(
        self, cfg: ConverterPRUConfig, cal: PruCalibration, storage_cfg: StoragePRUConfig
    ) -> None:
        self._cal: PruCalibration = cal
        self._cfg: ConverterPRUConfig = cfg

        self.storage = VirtualStorageModelPRU(storage_cfg)

        # simplifications for python
        self.R_input_kOhm = float(self._cfg.R_input_kOhm_n22) / 2**22

        # boost internal state
        self.V_input_uV: float = 0.0
        self.P_inp_fW: float = 0.0
        self.P_out_fW: float = 0.0
        self.interval_startup_disabled_drain_n: int = self._cfg.interval_startup_delay_drain_n

        # container for the stored energy
        self.V_mid_uV: float = self.storage.calc_V_OC_uV()

        # buck internal state
        self.enable_storage: bool = (int(self._cfg.converter_mode) & 0b0001) > 0
        self.enable_boost: bool = (int(self._cfg.converter_mode) & 0b0010) > 0
        self.enable_buck: bool = (int(self._cfg.converter_mode) & 0b0100) > 0
        self.enable_log_mid: bool = (int(self._cfg.converter_mode) & 0b1000) > 0
        # back-channel to hrv
        self.feedback_to_hrv: bool = (int(self._cfg.converter_mode) & 0b1_0000) > 0
        self.V_input_request_uV: int = int(self.V_mid_uV)

        self.V_out_dac_uV: float = self._cfg.V_output_uV
        self.V_out_dac_raw: int = self._cal.conv_uV_to_dac_raw(self._cfg.V_output_uV)
        self.power_good: bool = True

        # prepare hysteresis-thresholds
        self.dV_mid_enable_output_uV: float = self._cfg.dV_mid_enable_output_uV
        self.V_mid_enable_output_threshold_uV: float = self._cfg.V_mid_enable_output_threshold_uV
        self.V_mid_disable_output_threshold_uV: float = self._cfg.V_mid_disable_output_threshold_uV

        self.V_mid_enable_output_threshold_uV = max(
            self.dV_mid_enable_output_uV, self.V_mid_enable_output_threshold_uV
        )

        # pulled from update_states_and_output() due to easier static init
        self.sample_count: int = 0xFFFFFFF0
        self.is_outputting: bool = False
        self.vsource_skip_gpio_logging: bool = False

    def calc_inp_power(self, input_voltage_uV: float, input_current_nA: float) -> int:
        # Next 2 lines are Python-specific (model unsigned int)
        input_voltage_uV = max(0.0, input_voltage_uV)
        input_current_nA = max(0.0, input_current_nA)

        # Input diode
        if input_voltage_uV > self._cfg.V_input_drop_uV:
            input_voltage_uV -= self._cfg.V_input_drop_uV
        else:
            input_voltage_uV = 0.0

        input_voltage_uV = min(input_voltage_uV, self._cfg.V_input_max_uV)

        input_current_nA = min(input_current_nA, self._cfg.I_input_max_nA)

        self.V_input_uV = input_voltage_uV

        if self.enable_boost:
            if input_voltage_uV < self._cfg.V_input_boost_threshold_uV:
                input_voltage_uV = 0.0
            # TODO: vdrop in case of v_input > v_storage (non-boost)
        elif self.enable_storage:
            # no boost, but cap, for i.e. diode+cap (+resistor)
            V_diff_uV = (
                (input_voltage_uV - self.V_mid_uV) if (input_voltage_uV >= self.V_mid_uV) else 0
            )
            V_res_drop_uV = input_current_nA * self.R_input_kOhm
            if V_res_drop_uV > V_diff_uV:
                input_voltage_uV = self.V_mid_uV
            else:
                input_voltage_uV -= V_res_drop_uV

            # IF input==ivcurve request new CV
            if self.feedback_to_hrv:
                self.V_input_request_uV = self.V_mid_uV + V_res_drop_uV + self._cfg.V_input_drop_uV
            elif input_voltage_uV < self.V_mid_uV:
                # without feedback there is no usable energy here
                input_voltage_uV = 0
        else:
            # direct connection
            # modifying V_mid here is not clean, but simpler
            # -> V_mid is needed in calc_out, before cap is updated
            self.V_mid_uV = input_voltage_uV
            input_voltage_uV = 0.0
            # ⤷ input will not be evaluated

        if self.enable_boost:
            eta_inp = self.get_input_efficiency(input_voltage_uV, input_current_nA)
        else:
            eta_inp = 1.0

        self.P_inp_fW = eta_inp * input_voltage_uV * input_current_nA
        return round(self.P_inp_fW)  # Python-specific, added for easier testing

    def calc_out_power(self, current_adc_raw: int) -> int:
        # Next 2 lines are Python-specific (model unsigned int)
        current_adc_raw = max(0, current_adc_raw)
        current_adc_raw = min((2**18) - 1, current_adc_raw)

        # TODO: remove P_leak in C-Code
        I_out_nA = self._cal.conv_adc_raw_to_nA(current_adc_raw)
        if self.enable_buck:  # noqa: SIM108
            eta_inv_out = self.get_output_inv_efficiency(I_out_nA)
        else:
            eta_inv_out = 1.0

        self.P_out_fW = eta_inv_out * self.V_out_dac_uV * I_out_nA

        if self.interval_startup_disabled_drain_n > 0:
            self.interval_startup_disabled_drain_n -= 1
            self.P_out_fW = 0.0

        return round(self.P_out_fW)  # Python-specific, added for easier testing

    def update_cap_storage(self) -> int:
        if self.enable_storage:
            I_delta_nA = abs((self.P_inp_fW - self.P_out_fW) / self.V_mid_uV)
            is_charging: bool = self.P_inp_fW >= self.P_out_fW
            self.V_mid_uV = self.storage.step(2**4 * I_delta_nA, is_charging=is_charging)
        self.V_mid_uV = min(self.V_mid_uV, self._cfg.V_mid_max_uV)
        self.V_mid_uV = max(self.V_mid_uV, 1)
        return round(self.V_mid_uV)  # Python-specific, added for easier testing

    def update_states_and_output(self) -> int:
        self.sample_count += 1
        check_thresholds = self.sample_count >= self._cfg.interval_check_thresholds_n
        V_mid_uV_now = self.V_mid_uV
        # copy avoids not enabling pwr_good (due to large dV_enable_output_uV)

        if check_thresholds:
            self.sample_count = 0
            if self.is_outputting:
                if V_mid_uV_now < self.V_mid_disable_output_threshold_uV:
                    self.is_outputting = False
            elif V_mid_uV_now >= self.V_mid_enable_output_threshold_uV:
                self.is_outputting = True
                self.V_mid_uV -= self.dV_mid_enable_output_uV

        if check_thresholds or self._cfg.immediate_pwr_good_signal:
            # generate power-good-signal
            if self.power_good:
                if V_mid_uV_now <= self._cfg.V_pwr_good_disable_threshold_uV:
                    self.power_good = False
            elif V_mid_uV_now >= self._cfg.V_pwr_good_enable_threshold_uV:
                self.power_good = self.is_outputting
            # set batok pin to state ... TODO?

        if self.is_outputting or self.interval_startup_disabled_drain_n > 0:
            if (not self.enable_buck) or (
                self.V_mid_uV <= self._cfg.V_output_uV + self._cfg.V_buck_drop_uV
            ):
                if self.V_mid_uV > self._cfg.V_buck_drop_uV:
                    self.V_out_dac_uV = self.V_mid_uV - self._cfg.V_buck_drop_uV
                else:
                    self.V_out_dac_uV = 0.0
            else:
                self.V_out_dac_uV = self._cfg.V_output_uV
            self.V_out_dac_raw = self._cal.conv_uV_to_dac_raw(self.V_out_dac_uV)
        else:
            self.V_out_dac_uV = 0.0
            self.V_out_dac_raw = 0

        self.vsource_skip_gpio_logging = (
            self.V_out_dac_uV < self._cfg.V_output_log_gpio_threshold_uV
        )
        return self.V_out_dac_raw

    def get_input_efficiency(self, voltage_uV: float, current_nA: float) -> float:
        voltage_n = int(voltage_uV / (2**self._cfg.LUT_input_V_min_log2_uV))
        current_n = int(current_nA / (2**self._cfg.LUT_input_I_min_log2_nA))
        pos_v = int(voltage_n) if (voltage_n > 0) else 0  # V-Scale is Linear!
        pos_c = int(math.log2(current_n)) if (current_n > 0) else 0
        if pos_v >= LUT_SIZE:
            pos_v = LUT_SIZE - 1
        if pos_c >= LUT_SIZE:
            pos_c = LUT_SIZE - 1
        return self._cfg.LUT_inp_efficiency_n8[pos_v][pos_c] / (2**8)

    def get_output_inv_efficiency(self, current_nA: float) -> float:
        current_n = int(current_nA / (2**self._cfg.LUT_output_I_min_log2_nA))
        pos_c = int(math.log2(current_n)) if (current_n > 0) else 0
        if pos_c >= LUT_SIZE:
            pos_c = LUT_SIZE - 1
        return self._cfg.LUT_out_inv_efficiency_n4[pos_c] / (2**4)

    def set_P_input_fW(self, value: float) -> None:
        self.P_inp_fW = value

    def set_P_output_fW(self, value: float) -> None:
        self.P_out_fW = value

    def set_V_intermediate_uV(self, value: float) -> None:
        self.V_mid_uV = value

    def get_P_input_fW(self) -> int:
        return round(self.P_inp_fW)

    def get_P_output_fW(self) -> int:
        return round(self.P_out_fW)

    def get_V_intermediate_uV(self) -> int:
        return round(self.V_mid_uV)

    def get_V_intermediate_raw(self) -> int:
        return round(self._cal.conv_uV_to_dac_raw(self.V_mid_uV))

    def get_power_good(self) -> bool:
        return self.power_good

    def get_I_mid_out_nA(self) -> float:
        return self.P_out_fW / self.V_mid_uV

    def get_state_log_intermediate(self) -> bool:
        return self.enable_log_mid

    def get_state_log_gpio(self) -> bool:
        return self.vsource_skip_gpio_logging



"""Collection of tasks for all observers included in experiment."""

from pathlib import Path
from pathlib import PurePosixPath
from typing import TYPE_CHECKING
from typing import Annotated

from pydantic import Field
from pydantic import validate_call
from typing_extensions import Self

from shepherd_core.data_models.base.content import NameStr
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.experiment.experiment import Experiment
from shepherd_core.data_models.testbed.testbed import Testbed

from .observer_tasks import ObserverTasks

if TYPE_CHECKING:
    from collections.abc import Set as AbstractSet


class TestbedTasks(ShpModel):
    """Collection of tasks for all observers included in experiment."""

    name: NameStr
    observer_tasks: Annotated[list[ObserverTasks], Field(min_length=1, max_length=128)]

    @classmethod
    @validate_call
    def from_xp(cls, xp: Experiment, tb: Testbed | None = None) -> Self:
        if tb is None:
            # TODO: is tb-argument really needed? prob. not
            tb = Testbed()  # this will query the first (and only) entry of client

        tgt_ids = xp.get_target_ids()
        xp_folder = xp.folder_name()
        obs_tasks = [ObserverTasks.from_xp(xp, xp_folder, tb, _id) for _id in tgt_ids]
        return cls(
            name=xp.name,
            observer_tasks=obs_tasks,
        )

    def get_observer_tasks(self, observer: str) -> ObserverTasks | None:
        for tasks in self.observer_tasks:
            if observer == tasks.observer:
                return tasks
        return None

    def get_observers(self) -> set[str]:
        return {tasks.observer for tasks in self.observer_tasks}

    def get_output_paths(self) -> dict[str, Path]:
        # TODO: computed field preferred, but they don't work here, as
        #  - they are always stored in yaml despite "repr=False"
        #  - solution will shift to some kind of "result"-datatype that is combinable
        values: dict[str, Path] = {}
        for obt in self.observer_tasks:
            values = {**values, **obt.get_output_paths()}
        return values

    def is_contained(self) -> bool:
        """Limit paths to allowed directories."""
        paths_allowed: AbstractSet[PurePosixPath] = {
            PurePosixPath("/var/shepherd/"),
            PurePosixPath("/tmp/"),  # noqa: S108
        }
        return all(obt.is_contained(paths_allowed) for obt in self.observer_tasks)



"""meta-data representation of a testbed-component (physical object)."""

from datetime import datetime
from typing import Annotated
from typing import Any

from pydantic import Field
from pydantic import IPvAnyAddress
from pydantic import StringConstraints
from pydantic import model_validator
from typing_extensions import Self

from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.content import NameStr
from shepherd_core.data_models.base.content import SafeStr
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.testbed_client import tb_client

from .cape import Cape
from .cape import TargetPort
from .target import Target

MACStr = Annotated[
    str,
    StringConstraints(max_length=17, pattern=r"^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$"),
]


class Observer(ShpModel, title="Shepherd-Sheep"):
    """meta-data representation of a testbed-component (physical object)."""

    id: IdInt
    name: NameStr
    description: SafeStr
    comment: SafeStr | None = None

    ip: IPvAnyAddress
    mac: MACStr

    room: NameStr
    eth_port: NameStr

    latitude: Annotated[float, Field(ge=-90, le=90)] = 51.026573
    longitude: Annotated[float, Field(ge=-180, le=180)] = 13.723291
    """ ⤷ cfaed-floor"""

    active: bool = True
    cape: Cape | None = None
    target_a: Target | None = None
    target_b: Target | None = None

    created: datetime = Field(default_factory=datetime.now)
    alive_last: datetime | None = None

    def __str__(self) -> str:
        return self.name

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        values, _ = tb_client.try_completing_model(cls.__name__, values)
        return values

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        has_cape = self.cape is not None
        has_target = (self.target_a is not None) or (self.target_b is not None)
        if not has_cape and has_target:
            msg = f"Observer '{self.name}' is faulty -> has targets but no cape"
            raise ValueError(msg)
        return self

    def has_target(self, target_id: int) -> bool:
        case_a = (
            self.target_a is not None and target_id == self.target_a.id and self.target_a.active
        )
        case_b = (
            self.target_b is not None and target_id == self.target_b.id and self.target_b.active
        )
        return case_a or case_b

    def get_target_port(self, target_id: int) -> TargetPort:
        if self.has_target(target_id):
            if self.target_a is not None and target_id == self.target_a.id:
                return TargetPort.A
            if self.target_b is not None and target_id == self.target_b.id:
                return TargetPort.B
        msg = f"Target-ID {target_id} was not found in Observer '{self.name}'"
        raise KeyError(msg)

    def get_target(self, target_id: int) -> Target:
        if self.has_target(target_id):
            if self.target_a is not None and target_id == self.target_a.id:
                return self.target_a
            if self.target_b is not None and target_id == self.target_b.id:
                return self.target_b
        msg = f"Target-ID {target_id} was not found in Observer '{self.name}'"
        raise KeyError(msg)



"""Hardware related inventory model."""

from collections.abc import Sequence

from pydantic import ConfigDict
from typing_extensions import Self

from shepherd_core.data_models import ShpModel


class TargetInventory(ShpModel):
    """Hardware related inventory model."""

    cape: str | None = None
    targets: Sequence[str] = ()

    model_config = ConfigDict(str_min_length=0)

    @classmethod
    def collect(cls) -> Self:
        model_dict = {}

        return cls(**model_dict)



"""Simulator for the virtual storage models / algorithms."""

from collections.abc import Callable
from pathlib import Path

import numpy as np
from pydantic import PositiveFloat
from pydantic import validate_call

from shepherd_core import log

from .virtual_storage_model import ModelStorage


class StorageSimulator:
    """The simulator benchmarks a set of storage-models.

    - monitors cell-current and voltage, open circuit voltage, state of charge and time
    - takes config with a list of storage-models and timebase
    - runs with a total step-count as config and a current-providing function
        taking time, cell-voltage and SoC as arguments

    The recorded data can be visualized by generating plots.
    """

    def __init__(self, models: list[ModelStorage], dt_s: PositiveFloat) -> None:
        self.models = models
        self.dt_s = dt_s
        for model in self.models:
            if self.dt_s != model.dt_s:
                raise ValueError("timebase on models do not match")
        self.t_s: np.ndarray | None = None

        # models return V_cell, SoC_eff, V_OC
        self.I_input: np.ndarray | None = None
        self.V_OC: np.ndarray | None = None
        self.V_cell: np.ndarray | None = None
        self.SoC: np.ndarray | None = None
        self.SoC_eff: np.ndarray | None = None

    @validate_call
    def run(self, fn: Callable, duration_s: PositiveFloat) -> None:
        self.t_s = np.arange(0, duration_s, self.dt_s)
        self.I_input = np.zeros((len(self.models), self.t_s.shape[0]))
        self.V_OC = np.zeros((len(self.models), self.t_s.shape[0]))
        self.V_cell = np.zeros((len(self.models), self.t_s.shape[0]))
        self.SoC = np.zeros((len(self.models), self.t_s.shape[0]))
        self.SoC_eff = np.zeros((len(self.models), self.t_s.shape[0]))
        for i, model in enumerate(self.models):
            SoC = 1.0
            V_cell = 0.0
            for j, t_s in enumerate(self.t_s):
                I_charge = fn(t_s, SoC, V_cell)
                V_OC, V_cell, SoC, SoC_eff = model.step(I_charge)
                self.I_input[i, j] = I_charge
                self.V_OC[i, j] = V_OC
                self.V_cell[i, j] = V_cell
                self.SoC[i, j] = SoC
                self.SoC_eff[i, j] = SoC_eff

    @validate_call
    def plot(self, path: Path, title: str, *, plot_delta_v: bool = False) -> None:
        try:
            # keep dependencies low
            from matplotlib import pyplot as plt  # noqa: PLC0415
        except ImportError:
            log.warning("Matplotlib not installed, plotting of results disabled")
            return

        offset = 1 if plot_delta_v else 0
        fig, axs = plt.subplots(4 + offset, 1, sharex="all", figsize=(10, 2 * 6), layout="tight")
        axs[0].set_title(title)
        axs[0].set_ylabel("State of Charge [n]")
        # ⤷ Note: SoC-eff is also available, but unused
        axs[0].grid(visible=True)
        axs[1].set_ylabel("Open-circuit voltage [V]")
        axs[1].grid(visible=True)
        axs[2].set_ylabel("Cell voltage [V]")
        axs[2].grid(visible=True)
        if plot_delta_v:
            axs[3].set_ylabel("Cell voltage delta [V]")
            axs[3].grid(visible=True)
        axs[3 + offset].set_ylabel("Charge current [A]")
        axs[3 + offset].set_xlabel("time [s]")
        axs[3 + offset].grid(visible=True)

        for i, model in enumerate(self.models):
            axs[0].plot(
                self.t_s, self.SoC[i], label=f"{type(model).__name__} {model.cfg.name}", alpha=0.7
            )
            axs[1].plot(self.t_s, self.V_OC[i], label=type(model).__name__, alpha=0.7)
            axs[2].plot(self.t_s, self.V_cell[i], label=type(model).__name__, alpha=0.7)
            if plot_delta_v:  # assumes that timestamps are identical
                axs[3].plot(
                    self.t_s,
                    [v - ref for v, ref in zip(self.V_cell[i], self.V_cell[0], strict=False)],
                    label=type(model).__name__,
                    alpha=0.7,
                )
            axs[3 + offset].plot(self.t_s, self.I_input[i], label=type(model).__name__, alpha=0.7)
        axs[0].legend()
        plt.savefig(path / f"{title}.png")
        plt.close(fig)
        plt.clf()



"""Module for testbed-related data-models.

These models import externally from: /base
"""

from .cape import Cape
from .cape import TargetPort
from .gpio import GPIO
from .gpio import Direction
from .mcu import MCU
from .mcu import ProgrammerProtocol
from .observer import MACStr
from .observer import Observer
from .target import IdInt16
from .target import Target
from .testbed import Testbed

__all__ = [
    "GPIO",
    "MCU",
    "Cape",
    "Direction",
    "IdInt16",
    "MACStr",
    "Observer",
    "ProgrammerProtocol",
    "Target",
    "TargetPort",
    "Testbed",
]



"""Config for a Task programming the selected target."""

from collections.abc import Set as AbstractSet
from pathlib import Path
from pathlib import PurePosixPath
from typing import Annotated

from pydantic import Field
from pydantic import model_validator
from pydantic import validate_call
from typing_extensions import Self

from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.content import SafeStr
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.content.firmware import suffix_to_DType
from shepherd_core.data_models.content.firmware_datatype import FirmwareDType
from shepherd_core.data_models.experiment.experiment import Experiment
from shepherd_core.data_models.testbed.cape import TargetPort
from shepherd_core.data_models.testbed.mcu import ProgrammerProtocol
from shepherd_core.data_models.testbed.target import MCUPort
from shepherd_core.data_models.testbed.testbed import Testbed

from .helper_paths import path_posix


class ProgrammingTask(ShpModel):
    """Config for a Task programming the selected target."""

    firmware_file: Path
    target_port: TargetPort = TargetPort.A
    mcu_port: MCUPort = 1
    mcu_type: SafeStr
    """ ⤷ must be either "nrf52" or "msp430" ATM, TODO: clean xp to tasks"""
    voltage: Annotated[float, Field(ge=1, lt=5)] = 3
    datarate: Annotated[int, Field(gt=0, le=1_000_000)] = 200_000
    protocol: ProgrammerProtocol
    # TODO: eradicate - should not exist. derive protocol from mcu_type

    simulate: bool = False

    verbose: Annotated[int, Field(ge=0, le=4)] = 2
    """ ⤷ 0=Errors, 1=Warnings, 2=Info, 3=Debug"""

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        d_type = suffix_to_DType.get(self.firmware_file.suffix.lower())
        if d_type != FirmwareDType.base64_hex:
            msg = (f"Firmware is not intel-hex ({self.firmware_file.as_posix()})",)
            raise ValueError(msg)
        return self

    @classmethod
    @validate_call
    def from_xp(
        cls,
        xp: Experiment,
        tb: Testbed,
        tgt_id: IdInt,
        mcu_port: MCUPort,
        fw_path: Path,
    ) -> Self | None:
        obs = tb.get_observer(tgt_id)
        tgt_cfg = xp.get_target_config(tgt_id)

        fw = tgt_cfg.firmware1 if mcu_port == 1 else tgt_cfg.firmware2
        if fw is None:
            return None

        return cls(
            firmware_file=path_posix(fw_path),
            target_port=obs.get_target_port(tgt_id),
            mcu_port=mcu_port,
            mcu_type=fw.mcu.name,
            voltage=fw.mcu.prog_voltage,
            datarate=fw.mcu.prog_datarate,
            protocol=fw.mcu.prog_protocol,
        )

    def is_contained(self, paths: AbstractSet[PurePosixPath]) -> bool:
        """Limit paths to allowed directories."""
        return any(self.firmware_file.is_relative_to(path) for path in paths)



"""Models for the process of calibration a device by measurements."""

from typing import Annotated

import numpy as np
from pydantic import Field
from pydantic import PositiveFloat
from pydantic import validate_call

from .calibration import CalibrationCape
from .calibration import CalibrationEmulator
from .calibration import CalibrationHarvester
from .calibration import CalibrationPair
from .calibration import CapeData
from .shepherd import ShpModel


class CalMeasurementPair(ShpModel):
    """Value-container for a calibration-measurement."""

    shepherd_raw: PositiveFloat
    reference_si: float = 0


CalMeasPairs = Annotated[list[CalMeasurementPair], Field(min_length=2)]


@validate_call
def meas_to_cal(data: CalMeasPairs, component: str) -> CalibrationPair:
    """Convert values from calibration-measurement to the calibration itself."""
    x = np.empty(len(data))
    y = np.empty(len(data))
    for i, pair in enumerate(data):
        x[i] = pair.shepherd_raw
        y[i] = pair.reference_si

    model = np.polyfit(x, y, 1)
    offset: float = float(model[1])
    gain: float = float(model[0])

    # r-squared, Pearson correlation coefficient
    _p = np.poly1d(model)
    yhat = _p(x)
    ybar: float = np.sum(y) / len(y)
    ssreg: float = np.sum((yhat - ybar) ** 2)
    sstot: float = np.sum((y - ybar) ** 2)
    rval: float = ssreg / sstot

    if rval < 0.999:
        msg = (
            "Calibration faulty -> Correlation coefficient "
            f"(rvalue) = {rval}:.6f is too low for '{component}'"
        )
        raise ValueError(msg)
    return CalibrationPair(offset=offset, gain=gain)


class CalMeasurementHarvester(ShpModel):
    """Container for the values of the calibration-measurement."""

    dac_V_Hrv: CalMeasPairs
    dac_V_Sim: CalMeasPairs
    adc_V_Sense: CalMeasPairs
    adc_C_Hrv: CalMeasPairs

    def to_cal(self) -> CalibrationHarvester:
        dv = self.model_dump()
        dcal = CalibrationHarvester().model_dump()
        for key in dv:
            dcal[key] = meas_to_cal(self[key], f"hrv_{key}")
        return CalibrationHarvester(**dcal)


class CalMeasurementEmulator(ShpModel):
    """Container for the values of the calibration-measurement."""

    dac_V_A: CalMeasPairs  # TODO: why not V_dac_A or V_dac_a
    dac_V_B: CalMeasPairs
    adc_C_A: CalMeasPairs
    adc_C_B: CalMeasPairs

    def to_cal(self) -> CalibrationEmulator:
        dv = self.model_dump()
        dcal = CalibrationEmulator().model_dump()
        for key in dv:
            dcal[key] = meas_to_cal(self[key], f"emu_{key}")
        return CalibrationEmulator(**dcal)


class CalMeasurementCape(ShpModel):
    """Container for the values of the calibration-measurement."""

    cape: CapeData | None = None
    host: str | None = None

    harvester: CalMeasurementHarvester | None = None
    emulator: CalMeasurementEmulator | None = None

    def to_cal(self) -> CalibrationCape:
        dv = self.model_dump()
        dcal = CalibrationCape().model_dump()
        # TODO: is it helpful to default wrong / missing values?
        for key, value in dv.items():
            if key in {"harvester", "emulator"}:
                if value is not None:
                    dcal[key] = self[key].to_cal()
            else:
                dcal[key] = self[key]

        return CalibrationCape(**dcal)



"""meta-data representation of a testbed-component (physical object)."""

from datetime import timedelta
from pathlib import Path
from typing import Annotated
from typing import Any

from pydantic import Field
from pydantic import HttpUrl
from pydantic import model_validator
from typing_extensions import Self

from shepherd_core.config import config
from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.content import NameStr
from shepherd_core.data_models.base.content import SafeStr
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.testbed_client import tb_client

from .observer import Observer

duration_5min = timedelta(minutes=5)


class Testbed(ShpModel):
    """meta-data representation of a testbed-component (physical object)."""

    id: IdInt
    name: NameStr
    description: SafeStr
    comment: SafeStr | None = None

    url: HttpUrl | None = None

    observers: Annotated[list[Observer], Field(min_length=1, max_length=128)]

    shared_storage: bool = True
    data_on_server: Path
    data_on_observer: Path
    """ ⤷ storage layout: root_path/content_type/group/owner/[object]"""
    # TODO: we might need individual paths for experiments & content

    prep_duration: timedelta = duration_5min
    # TODO: one BBone is currently time-keeper

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        # allow instantiating an empty Testbed, take default in config
        if len(values) == 0:
            values = {"name": config.TESTBED}

        values, _ = tb_client.try_completing_model(cls.__name__, values)
        return values

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        observers = []
        ips = []
        macs = []
        capes = []
        targets = []
        eth_ports = []
        for _obs in self.observers:
            observers.append(_obs.id)
            ips.append(_obs.ip)
            macs.append(_obs.mac)
            if _obs.cape is not None:
                capes.append(_obs.cape)
            if _obs.target_a is not None:
                targets.append(_obs.target_a)
            if _obs.target_b is not None:
                targets.append(_obs.target_b)
            eth_ports.append(_obs.eth_port)
        if len(observers) > len(set(observers)):
            raise ValueError("Observers used more than once in Testbed")
        if len(ips) > len(set(ips)):
            raise ValueError("Observer-IP used more than once in Testbed")
        if len(macs) > len(set(macs)):
            raise ValueError("Observers-MAC-Address used more than once in Testbed")
        if len(capes) > len(set(capes)):
            raise ValueError("Cape used more than once in Testbed")
        if len(targets) > len(set(targets)):
            raise ValueError("Target used more than once in Testbed")
        if len(eth_ports) > len(set(eth_ports)):
            raise ValueError("Observers-Ethernet-Port used more than once in Testbed")
        if self.prep_duration.total_seconds() < 0:
            raise ValueError("Task-Duration can't be negative.")
        if not self.shared_storage:
            raise ValueError("Only shared-storage-option is implemented")
        return self

    def get_observer(self, target_id: int) -> Observer:
        for _observer in self.observers:
            if not _observer.active or not _observer.cape.active:
                # skip decommissioned setups
                continue
            if _observer.has_target(target_id):
                return _observer
        msg = f"Target-ID {target_id} was not found in Testbed '{self.name}'"
        raise ValueError(msg)



r"""Helper FN to avoid unwanted behavior.

On windows Path("\xyz") gets transformed to "/xyz", but not on linux.
When sending an experiment via fastapi, this bug gets triggered.
"""

from pathlib import Path


def path_posix(path: Path) -> Path:
    r"""Help Linux to get from "\xyz" to "/xyz".

    This isn't a problem on windows and gets triggered when sending XP via fastapi.
    """
    return Path(path.as_posix().replace("\\", "/"))



from pathlib import Path

import pytest
from pydantic import ValidationError
from shepherd_core.data_models.content import EnergyDType
from shepherd_core.data_models.content import EnergyEnvironment
from shepherd_core.data_models.content import Firmware
from shepherd_core.data_models.content import FirmwareDType
from shepherd_core.data_models.content import VirtualHarvesterConfig
from shepherd_core.data_models.content import VirtualSourceConfig
from shepherd_core.data_models.content.virtual_source_config import ConverterPRUConfig
from shepherd_core.data_models.content.virtual_storage_config import VirtualStorageConfig
from shepherd_core.data_models.testbed import MCU

from shepherd_core import fw_tools

from .conftest import files_elf


def test_content_model_ee_min1() -> None:
    EnergyEnvironment(
        id=9999,
        name="some",
        data_path="./file",
        data_type="isc_voc",
        duration=1,
        energy_Ws=0.1,
        owner="jane",
        group="wayne",
    )


def test_content_model_ee_min2() -> None:
    EnergyEnvironment(
        id="98765",
        name="some",
        data_path="./file",
        data_type=EnergyDType.ivcurve,
        duration=999,
        energy_Ws=3.1,
        owner="jane",
        group="wayne",
    )


def test_content_model_fw_faulty() -> None:
    with pytest.raises(ValidationError):
        Firmware(
            id=9999,
            name="dome",
            mcu=MCU(name="nRF52"),
            data="xyz",
            data_type=FirmwareDType.base64_hex,
            owner="Obelix",
            group="Gaul",
        )


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_content_model_fw_min(path_elf: Path, tmp_path: Path) -> None:
    path_hex = (tmp_path / (path_elf.stem + ".hex")).resolve()
    path_hex = fw_tools.elf_to_hex(path_elf, path_hex)
    Firmware(
        id=9999,
        name="dome",
        mcu=MCU(name="nRF52"),
        data=fw_tools.file_to_base64(path_hex),
        data_type=FirmwareDType.base64_hex,
        owner="Obelix",
        group="Gaul",
    )


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_content_model_fw_from_elf(path_elf: Path) -> None:
    Firmware.from_firmware(
        file=path_elf,
        name="dome",
        owner="Obelix",
        group="Gaul",
    )


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_content_model_fw_from_hex(path_elf: Path, tmp_path: Path) -> None:
    path_hex = (tmp_path / (path_elf.stem + ".hex")).resolve()
    path_hex = fw_tools.elf_to_hex(path_elf, path_hex)
    Firmware.from_firmware(
        file=path_hex,
        name="dome",
        owner="Obelix",
        group="Gaul",
    )


def test_content_model_fw_from_hex_failing(tmp_path: Path) -> None:
    path_hex = tmp_path / "some.hex"
    with path_hex.open("w") as fd:
        fd.write("something")
    with pytest.raises(ValueError):  # noqa: PT011
        Firmware.from_firmware(
            file=path_hex,
            name="dome",
            owner="Obelix",
            group="Gaul",
        )


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_content_model_fw_extract_elf_to_dir(path_elf: Path, tmp_path: Path) -> None:
    fw = Firmware.from_firmware(
        file=path_elf,
        name="dome",
        owner="Obelix",
        group="Gaul",
    )
    file = fw.extract_firmware(tmp_path)
    assert file.exists()
    assert file.is_file()


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_content_model_fw_extract_hex_to_dir(path_elf: Path, tmp_path: Path) -> None:
    path_hex = (tmp_path / (path_elf.stem + ".hex")).resolve()
    path_hex = fw_tools.elf_to_hex(path_elf, path_hex)
    fw = Firmware.from_firmware(
        file=path_hex,
        name="dome",
        owner="Obelix",
        group="Gaul",
    )
    file = fw.extract_firmware(tmp_path)
    assert file.exists()
    assert file.is_file()


@pytest.mark.parametrize("path_elf", files_elf)
def test_content_model_fw_extract_path_elf_to_dir(path_elf: Path, tmp_path: Path) -> None:
    assert path_elf.exists()
    fw = Firmware(
        data=path_elf,
        data_type=FirmwareDType.path_elf,
        mcu={"name": "MSP430FR"},
        name="dome",
        owner="Obelix",
        group="Gaul",
    )
    file = fw.extract_firmware(tmp_path)
    assert file.exists()
    assert file.is_file()


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_content_model_fw_extract_path_hex_to_dir(path_elf: Path, tmp_path: Path) -> None:
    path_hex = (tmp_path / (path_elf.stem + ".hex")).resolve()
    path_hex = fw_tools.elf_to_hex(path_elf, path_hex)
    assert path_hex.exists()
    fw = Firmware(
        data=path_hex,
        data_type=FirmwareDType.path_hex,
        mcu={"name": "MSP430FR"},
        name="dome",
        owner="Obelix",
        group="Gaul",
    )
    file = fw.extract_firmware(tmp_path)
    assert file.exists()
    assert file.is_file()


def test_content_model_hrv_min() -> None:
    hrv = VirtualHarvesterConfig(
        id=9999,
        name="whatever",
        owner="jane",
        group="wayne",
        algorithm="mppt_opt",
    )
    assert hrv.get_datatype() == "ivsample"


def test_content_model_hrv_neutral() -> None:
    with pytest.raises(ValidationError):
        _ = VirtualHarvesterConfig(name="neutral")


@pytest.mark.parametrize("name", ["iv110", "cv24", "mppt_voc", "mppt_po"])
def test_content_model_hrv_by_name(name: str) -> None:
    _ = VirtualHarvesterConfig(name=name)


@pytest.mark.parametrize("uid", [1103, 1200, 2102, 2204, 2205, 2206])
def test_content_model_hrv_by_id(uid: int) -> None:
    _ = VirtualHarvesterConfig(id=uid)


def test_content_model_hrv_steps() -> None:
    hrv = VirtualHarvesterConfig(
        name="ivsurface", voltage_min_mV=1000, voltage_max_mV=4000, samples_n=11
    )
    assert hrv.voltage_step_mV == 300


def test_content_model_hrv_faulty_voltage0() -> None:
    with pytest.raises(ValidationError):
        _ = VirtualHarvesterConfig(name="iv110", voltage_max_mV=5001)
    with pytest.raises(ValidationError):
        _ = VirtualHarvesterConfig(name="iv110", voltage_min_mV=-1)


def test_content_model_hrv_faulty_voltage1() -> None:
    with pytest.raises(ValidationError):
        _ = VirtualHarvesterConfig(name="iv110", voltage_min_mV=4001, voltage_max_mV=4000)


def test_content_model_hrv_faulty_voltage2() -> None:
    with pytest.raises(ValidationError):
        _ = VirtualHarvesterConfig(name="iv110", voltage_mV=4001, voltage_max_mV=4000)


def test_content_model_hrv_faulty_voltage3() -> None:
    with pytest.raises(ValidationError):
        _ = VirtualHarvesterConfig(name="iv110", voltage_mV=4000, voltage_min_mV=4001)


@pytest.mark.parametrize("name", ["ivsurface", "iv1000", "isc_voc"])
def test_content_model_hrv_faulty_emu(name: str) -> None:
    hrv = VirtualHarvesterConfig(name=name)
    with pytest.raises(ValidationError):
        _ = VirtualSourceConfig(name="dio_cap", harvester=hrv)


def test_content_model_src_min() -> None:
    VirtualSourceConfig(
        id=9999,
        name="new_src",
        owner="jane",
        group="wayne",
    )


def test_content_model_src_force_warning() -> None:
    src = VirtualSourceConfig(
        name="BQ25570",
        C_output_uF=200,
        storage=VirtualStorageConfig.capacitor(C_uF=100, V_rated=6.3),
    )
    ConverterPRUConfig.from_vsrc(src, dtype_in=EnergyDType.ivsample)
    # -> triggers warning currently


def test_content_model_src_force_other_hysteresis1() -> None:
    src = VirtualSourceConfig(
        name="BQ25570",
        V_intermediate_enable_output_threshold_mV=4000,
        V_intermediate_disable_output_threshold_mV=3999,
        V_output_mV=2000,
        V_buck_drop_mV=100,
    )
    ConverterPRUConfig.from_vsrc(src, dtype_in=EnergyDType.ivsample)


def test_content_model_src_force_other_hysteresis2() -> None:
    src = VirtualSourceConfig(
        name="BQ25570",
        V_intermediate_enable_output_threshold_mV=1000,
        V_intermediate_disable_output_threshold_mV=999,
        V_output_mV=2000,
        V_buck_drop_mV=100,
    )
    ConverterPRUConfig.from_vsrc(src, dtype_in=EnergyDType.ivsample)



"""this is ported py-version of the pru-code.

Goals:

- stay close to original code-base (fixed-point integer math)
- offer a comparison for the tests
- step 1 to a virtualization of emulation

NOTE1: DO NOT OPTIMIZE -> stay close to original c-code-base

Compromises:

- Py has to map the settings-list to internal vars -> is kernel-task

Expected deviations:

- lead charge ramp maxes out early on cell-voltage (max of V_uV_n8 is 16.78 V)

"""

from pydantic import PositiveFloat
from pydantic import validate_call

from shepherd_core import log
from shepherd_core.data_models.content.virtual_storage_config import LuT_SIZE
from shepherd_core.data_models.content.virtual_storage_config import LuT_SIZE_LOG
from shepherd_core.data_models.content.virtual_storage_config import StoragePRUConfig
from shepherd_core.data_models.content.virtual_storage_config import TIMESTEP_s_DEFAULT
from shepherd_core.data_models.content.virtual_storage_config import VirtualStorageConfig
from shepherd_core.data_models.content.virtual_storage_config import soc_t


class ModelStorage:
    """Abstract base class for storage models."""

    def step(self, I_charge_A: float) -> tuple[float, float, float, float]: ...


def u32s(i: float) -> int:
    """Guard to supervise calculated model-states."""
    if i >= 2**32:
        log.warning("u32-overflow")
    if i < 0:
        log.warning("u32-underflow")
    return int(min(max(i, 0), 2**32 - 1))


def u64s(i: float) -> int:
    """Guard to supervise calculated model-states."""
    if i >= 2**64:
        log.warning("u64-overflow")
    if i < 0:
        log.warning("u64-underflow")
    return int(min(max(i, 0), 2**64 - 1))


class VirtualStorageModelPRU:
    """Ported python version of the pru vStorage.

    This model should behave like ModelKiBaMSimple
    """

    SoC_MAX_1_n62: int = 2**62 - 1
    SoC_TO_POS_DIV: int = 2 ** (62 - LuT_SIZE_LOG)

    @validate_call
    def __init__(
        self,
        cfg: StoragePRUConfig,
        SoC_init: soc_t | None = None,
    ) -> None:
        self.cfg_pru = cfg
        # state
        SoC_1_n30: float = 2**30 * SoC_init if SoC_init is not None else self.cfg_pru.SoC_init_1_n30
        self.SoC_1_n62 = round(2**32 * SoC_1_n30)
        self.V_OC_uV_n8 = self.cfg_pru.LuT_VOC_uV_n8[self.pos_LuT(self.SoC_1_n62)]

    def pos_LuT(self, SoC_1_n62: int) -> int:
        pos = u32s(SoC_1_n62 // self.SoC_TO_POS_DIV)
        if pos >= LuT_SIZE:
            pos = LuT_SIZE - 1
        return pos

    def calc_V_OC_uV(self) -> int:
        pos_LuT = self.pos_LuT(self.SoC_1_n62)
        return round(self.cfg_pru.LuT_VOC_uV_n8[pos_LuT] // 2**8)

    def step(self, I_delta_nA_n4: float, *, is_charging: bool) -> float:
        """Calculate the battery SoC & cell-voltage after drawing a current over a time-step.

        Note: 3x u64 multiplications
        """
        dSoC_leak_1_n62 = u64s((self.V_OC_uV_n8 // 2**6) * self.cfg_pru.Constant_1_per_uV_n60)
        if self.SoC_1_n62 >= dSoC_leak_1_n62:
            self.SoC_1_n62 = u64s(self.SoC_1_n62 - dSoC_leak_1_n62)
        else:
            self.SoC_1_n62 = 0

        dSoC_1_n62 = u64s(I_delta_nA_n4 * self.cfg_pru.Constant_1_per_nA_n60 // (2**2))
        if is_charging:
            self.SoC_1_n62 = u64s(self.SoC_1_n62 + dSoC_1_n62)
            self.SoC_1_n62 = min(self.SoC_MAX_1_n62, self.SoC_1_n62)
        elif self.SoC_1_n62 > dSoC_1_n62:
            self.SoC_1_n62 = u64s(self.SoC_1_n62 - dSoC_1_n62)
        else:
            self.SoC_1_n62 = 0

        pos_LuT = self.pos_LuT(self.SoC_1_n62)
        self.V_OC_uV_n8 = self.cfg_pru.LuT_VOC_uV_n8[pos_LuT]
        # TODO: is interpolation possible?
        R_series_kOhm_n32 = self.cfg_pru.LuT_RSeries_kOhm_n32[pos_LuT]
        V_delta_uV_n8 = u32s(u64s(I_delta_nA_n4 * R_series_kOhm_n32) // 2**28)

        if is_charging:
            V_cell_uV_n8 = u32s(self.V_OC_uV_n8 + V_delta_uV_n8)
        elif self.V_OC_uV_n8 > V_delta_uV_n8:
            V_cell_uV_n8 = u32s(self.V_OC_uV_n8 - V_delta_uV_n8)
        else:
            V_cell_uV_n8 = 0

        if self.SoC_1_n62 == 0:
            return 0  # cell voltage breaks down
        return V_cell_uV_n8 // 2**8  # u32 uV


class VirtualStorageModel(VirtualStorageModelPRU, ModelStorage):
    """Higher level Model that can run on a coarser timebase.

    This model should behave like ModelKiBaMSimple
    """

    @validate_call
    def __init__(
        self,
        cfg: VirtualStorageConfig,
        SoC_init: soc_t | None = None,
        dt_s: PositiveFloat = TIMESTEP_s_DEFAULT,
        *,
        optimize_clamp: bool = True,
    ) -> None:
        # metadata for simulator
        self.cfg: VirtualStorageConfig = cfg
        self.dt_s = dt_s
        # prepare PRU-Model
        cfg_pru = StoragePRUConfig.from_vstorage(
            cfg, TIMESTEP_s_DEFAULT, optimize_clamp=optimize_clamp
        )
        super().__init__(cfg_pru, SoC_init=SoC_init)

        # just for simulation
        self.steps_per_frame = round(dt_s / TIMESTEP_s_DEFAULT)

    def step(self, I_charge_A: float) -> tuple[float, float, float, float]:
        """Slower outer step with step-size of simulation."""
        I_delta_nA_n4 = abs(2**4 * (1e9 * I_charge_A))
        is_charging = I_charge_A >= 0
        for _ in range(self.steps_per_frame - 1):
            super().step(I_delta_nA_n4, is_charging=is_charging)
        V_cell_uV = super().step(I_delta_nA_n4, is_charging=is_charging)
        # code below just for simulation
        V_OC = (1e-6 / 2**8) * self.V_OC_uV_n8
        V_cell = 1e-6 * V_cell_uV
        SoC = (1.0 / 2**62) * self.SoC_1_n62
        return V_OC, V_cell, SoC, SoC



"""parametrized math models for converting between raw- and si-values.

Contains some info about the hardware configuration on the shepherd
cape. Based on these values, one can derive the expected adc readings given
an input voltage/current or, for emulation, the expected voltage and current
given the digital code in the DAC.

currently configured for cape v2.4c

"""

# both current channels have a 0.1 % shunt resistance of
R_SHT: float = 2.0  # [ohm]
# the instrumentation amplifiers are configured for gain of
G_INST_AMP: int = 48  # [n]
# we use the ADC's internal reference with
V_REF_ADC: float = 4.096  # [V]
# range of current channels is
G_ADC_I: float = 1.25  # [gain / V_REF]
# range of voltage channels is
G_ADC_V: float = 1.25  # [gain / V_REF]
# bit resolution of ADC
M_ADC: int = 18  # [bit]
# DACs use internal reference with
V_REF_DAC: float = 2.5  # [V]
# gain of DAC is set to
G_DAC: int = 2  # [n]
# bit resolution of DAC
M_DAC: int = 16  # [bit]
# DERIVED VARIABLES
RAW_MAX_ADC: int = 2**M_ADC - 1
RAW_MAX_DAC: int = 2**M_DAC - 1


def adc_current_to_raw(current: float, *, limited: bool = True) -> int:
    """Convert back a current [A] to raw ADC value."""
    # voltage on input of adc
    val_adc = G_INST_AMP * R_SHT * current
    # digital value according to ADC gain
    val_raw = int(val_adc * (2**M_ADC) / (G_ADC_I * V_REF_ADC))
    return min(max(val_raw, 0), 2**M_ADC - 1) if limited else val_raw


def adc_raw_to_current(value: int, *, limited: bool = True) -> float:
    """Convert a raw ADC value to a current [A]."""
    if limited:
        value = min(max(value, 0), 2**M_ADC - 1)
    # voltage on input of adc
    val_adc = float(value) * (G_ADC_I * V_REF_ADC) / (2**M_ADC)
    # current according to adc value
    return val_adc / (R_SHT * G_INST_AMP)


def adc_voltage_to_raw(voltage: float, *, limited: bool = True) -> int:
    """Convert back a voltage [V] to raw ADC value."""
    # digital value according to ADC gain
    val_raw = int(voltage * (2**M_ADC) / (G_ADC_V * V_REF_ADC))
    return min(max(val_raw, 0), 2**M_ADC - 1) if limited else val_raw


def adc_raw_to_voltage(value: int, *, limited: bool = True) -> float:
    """Convert a raw ADC value to a voltage [V]."""
    if limited:
        value = min(max(value, 0), 2**M_ADC - 1)
    # voltage according to ADC value
    return float(value) * (G_ADC_V * V_REF_ADC) / (2**M_ADC)


def dac_raw_to_voltage(value: int, *, limited: bool = True) -> float:
    """Convert back a raw DAC value to a voltage [V]."""
    if limited:
        value = min(max(value, 0), 2**M_DAC - 1)
    return float(value) * (V_REF_DAC * G_DAC) / (2**M_DAC)


def dac_voltage_to_raw(voltage: float, *, limited: bool = True) -> int:
    """Convert a voltage [V] to raw DAC value."""
    val_raw = int(voltage * (2**M_DAC) / (V_REF_DAC * G_DAC))
    return min(max(val_raw, 0), 2**M_DAC - 1) if limited else val_raw



"""Collection of tasks for selected observer included in experiment."""

from collections.abc import Set as AbstractSet
from datetime import datetime
from pathlib import Path
from pathlib import PurePosixPath
from typing import Annotated

from pydantic import validate_call
from typing_extensions import Self
from typing_extensions import deprecated

from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.content import NameStr
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.experiment.experiment import Experiment
from shepherd_core.data_models.testbed.testbed import Testbed

from .emulation import EmulationTask
from .firmware_mod import FirmwareModTask
from .helper_paths import path_posix
from .programming import ProgrammingTask


class ObserverTasks(ShpModel):
    """Collection of tasks for selected observer included in experiment."""

    observer: NameStr

    # PRE PROCESS
    time_prep: datetime | None = None  # TODO: currently not used
    root_path: Path

    # fw mod, store as hex-file and program
    fw1_mod: FirmwareModTask | None = None
    fw2_mod: FirmwareModTask | None = None
    fw1_prog: ProgrammingTask | None = None
    fw2_prog: ProgrammingTask | None = None

    # MAIN PROCESS
    emulation: EmulationTask | None = None

    # deprecations, TODO: remove before public release
    owner_id: Annotated[IdInt | None, deprecated("not needed anymore")] = None
    abort_on_error: Annotated[bool, deprecated("has no effect")] = False

    # post_copy / cleanup, Todo: could also just intake emuTask
    #  - delete firmwares
    #  - decode uart
    #  - downsample
    #  - zip it

    @classmethod
    @validate_call
    def from_xp(cls, xp: Experiment, xp_folder: str | None, tb: Testbed, tgt_id: IdInt) -> Self:
        if not tb.shared_storage:
            raise ValueError("Implementation currently relies on shared storage!")

        obs = tb.get_observer(tgt_id)
        if xp_folder is None:
            xp_folder = xp.folder_name()  # moved a layer up for consistent naming
        root_path = tb.data_on_observer / "experiments" / xp_folder
        fw_paths = [root_path / f"fw{_i}_{obs.name}.hex" for _i in [1, 2]]

        return cls(
            observer=obs.name,
            # time_prep=
            root_path=path_posix(root_path),
            fw1_mod=FirmwareModTask.from_xp(xp, tb, tgt_id, 1, fw_paths[0]),
            fw2_mod=FirmwareModTask.from_xp(xp, tb, tgt_id, 2, fw_paths[1]),
            fw1_prog=ProgrammingTask.from_xp(xp, tb, tgt_id, 1, fw_paths[0]),
            fw2_prog=ProgrammingTask.from_xp(xp, tb, tgt_id, 2, fw_paths[1]),
            emulation=EmulationTask.from_xp(xp, tb, tgt_id, root_path),
        )

    def get_tasks(self) -> list[ShpModel]:
        task_names = ["fw1_mod", "fw2_mod", "fw1_prog", "fw2_prog", "emulation"]
        tasks = []

        for task_name in task_names:
            if not hasattr(self, task_name):
                continue
            task = getattr(self, task_name)
            if task is None:
                continue
            tasks.append(task)
        return tasks

    def get_output_paths(self) -> dict[str, Path]:
        values: dict[str, Path] = {}
        if isinstance(self.emulation, EmulationTask):
            if self.emulation.output_path is None:
                raise ValueError("Emu-Task should have a valid output-path")
            values[self.observer] = self.emulation.output_path
        return values

    def is_contained(self, paths: AbstractSet[PurePosixPath]) -> bool:
        """Limit paths to allowed directories."""
        all_ok = any(self.root_path.is_relative_to(path) for path in paths)
        all_ok &= self.fw1_mod.is_contained(paths)
        all_ok &= self.fw2_mod.is_contained(paths)
        all_ok &= self.fw1_prog.is_contained(paths)
        all_ok &= self.fw2_prog.is_contained(paths)
        all_ok &= self.emulation.is_contained(paths)
        return all_ok



"""Create a set of static artificial energy environments."""

from itertools import product
from pathlib import Path

import numpy as np
from shepherd_core.logger import log
from tqdm import trange

from shepherd_core import Reader as ShpReader
from shepherd_core import Writer as ShpWriter

# Config
voltages_V = [4.0, 2.0]
currents_A = [2e-3, 1e-3]
duration_s = 10
repetitions = 1

path_here = Path(__file__).parent.absolute()

for _v, _c in product(voltages_V, currents_A):
    v_str = f"{round(_v * 1000)}mV"
    c_str = f"{round(_c * 1000)}mA"
    t_str = f"{round(duration_s * repetitions)}s"
    name = f"eenv_static_{v_str}_{c_str}_{t_str}"
    file_path = path_here / f"{name}.h5"

    if file_path.exists():
        log.info("File exists, will skip: %s", file_path.name)
    else:
        with ShpWriter(file_path) as file:
            file.store_hostname("artificial")
            # values in SI units
            timestamp_vector = np.arange(0.0, duration_s, file.sample_interval_ns / 1e9)
            voltage_vector = np.linspace(_v, _v, int(file.samplerate_sps * duration_s))
            current_vector = np.linspace(_c, _c, int(file.samplerate_sps * duration_s))

            for idx in trange(repetitions, desc="generate", leave=False):
                timestamps = idx * duration_s + timestamp_vector
                file.append_iv_data_si(timestamps, voltage_vector, current_vector)

    with ShpReader(file_path) as file:
        file.save_metadata()



"""Content-Converters for firmwares."""

import base64
import hashlib
import shutil
from pathlib import Path

import zstandard as zstd
from pydantic import validate_call

from shepherd_core.data_models.content.firmware_datatype import FirmwareDType

from .converter_elf import elf_to_hex
from .validation import is_elf
from .validation import is_hex


@validate_call
def firmware_to_hex(file_path: Path) -> Path:
    """Convert ELF-Files to HEX.

    Generic converter that handles ELF & HEX.
    """
    if not file_path.is_file():
        raise FileNotFoundError("Fn needs an existing file as input")
    if is_elf(file_path):
        return elf_to_hex(file_path)
    if is_hex(file_path):
        return file_path
    msg = (f"FW2Hex: unknown file '{file_path.name}', it should be ELF or HEX",)
    raise FileNotFoundError(msg)


@validate_call
def file_to_base64(file_path: Path) -> str:
    """Compress and encode content of file.

    - base64 adds ~33 % overhead
    - zstd compression reduces to ~ 1:3
    """
    if not file_path.is_file():
        raise ValueError("Fn needs an existing file as input")
    with file_path.resolve().open("rb") as file:
        file_content = file.read()
    file_cmpress = zstd.ZstdCompressor(level=20).compress(file_content)
    return base64.b64encode(file_cmpress).decode("ascii")


@validate_call
def base64_to_file(content: str, file_path: Path) -> None:
    """DeCompress and decode Content of file.

    - base64 adds ~33 % overhead
    - zstd compression reduces to ~ 1:3
    """
    file_cmpress = base64.b64decode(content)
    file_content = zstd.ZstdDecompressor().decompress(file_cmpress)
    if not file_path.parent.exists():
        file_path.parent.mkdir(parents=True)
    with file_path.resolve().open("wb") as file:
        file.write(file_content)


@validate_call
def file_to_hash(file_path: Path) -> str:
    """Convert file-content to hash-value."""
    if not file_path.is_file():
        raise ValueError("Fn needs an existing file as input")
    with file_path.resolve().open("rb") as file:
        file_content = file.read()
    return hashlib.sha3_224(file_content).hexdigest()


@validate_call
def base64_to_hash(content: str) -> str:
    """Convert base64-content to hash-value."""
    file_cmpress = base64.b64decode(content)
    file_content = zstd.ZstdDecompressor().decompress(file_cmpress)
    return hashlib.sha3_224(file_content).hexdigest()


@validate_call
def extract_firmware(data: str | Path, data_type: FirmwareDType, file_path: Path) -> Path:
    """Make embedded firmware-data usable in filesystem.

    - base64-string will be transformed to file
    - if data is a path the file will be copied to the destination.
    """
    if data_type == FirmwareDType.base64_elf and isinstance(data, str):
        file = file_path.with_suffix(".elf")
        base64_to_file(data, file)
    elif data_type == FirmwareDType.base64_hex and isinstance(data, str):
        file = file_path.with_suffix(".hex")
        base64_to_file(data, file)
    elif isinstance(data, (Path, str)):
        if data_type == FirmwareDType.path_elf:
            file = file_path.with_suffix(".elf")
        elif data_type == FirmwareDType.path_hex:
            file = file_path.with_suffix(".hex")
        else:
            msg = "FW-Extraction failed due to unknown datatype '{data_type}'"
            raise ValueError(msg)
        if not file.parent.exists():
            file.parent.mkdir(parents=True)
        shutil.copy(data, file)
    else:
        msg = f"FW-Extraction failed due to unknown data-type '{type(data)}'"
        raise ValueError(msg)
    return file



"""Shepherds base-model that brings a lot of default functionality."""

import hashlib
import pathlib
import pickle
from collections.abc import Generator
from datetime import timedelta
from ipaddress import IPv4Address
from pathlib import Path
from typing import Any
from uuid import UUID

import yaml
from pydantic import BaseModel
from pydantic import ConfigDict
from typing_extensions import Self
from yaml import Node
from yaml import SafeDumper

from .timezone import local_now
from .wrapper import Wrapper


def path2str(
    dumper: SafeDumper, data: pathlib.Path | pathlib.WindowsPath | pathlib.PosixPath
) -> Node:
    """Add a yaml-representation for a specific datatype."""
    return dumper.represent_scalar("tag:yaml.org,2002:str", str(data.as_posix()))


def time2int(dumper: SafeDumper, data: timedelta) -> Node:
    """Add a yaml-representation for a specific datatype."""
    return dumper.represent_scalar("tag:yaml.org,2002:int", str(int(data.total_seconds())))


def generic2str(dumper: SafeDumper, data: Any) -> Node:
    """Add a yaml-representation for a specific datatype."""
    return dumper.represent_scalar("tag:yaml.org,2002:str", str(data))


yaml.add_representer(pathlib.PosixPath, path2str, SafeDumper)
yaml.add_representer(pathlib.WindowsPath, path2str, SafeDumper)
yaml.add_representer(pathlib.Path, path2str, SafeDumper)
yaml.add_representer(timedelta, time2int, SafeDumper)
yaml.add_representer(IPv4Address, generic2str, SafeDumper)
yaml.add_representer(UUID, generic2str, SafeDumper)


def path_to_str(old: dict) -> dict:
    r"""Allow platform-independent pickling of ShpModel.

    Helper Fn
    Posix-Paths (/xyz/abc) in WindowsPath gets converted to \\xyz\\abc when exported
    intended usage: pickle.dump(path_to_str(model.model_dump()))
    """
    new: dict = {}
    for key, value in old.items():
        if isinstance(value, Path):
            new[key] = value.as_posix().replace("\\", "/")
        elif isinstance(value, dict):
            new[key] = path_to_str(value)
        else:
            new[key] = value
    return new


class ShpModel(BaseModel):
    """Pre-configured Pydantic Base-Model (specifically for shepherd).

    Inheritable Features:
    - constant / frozen, hashable .get_hash()
    - safe / limited custom types
    - string-representation str(ShpModel)
    - accessible as class (model.var) and dict (model[var])
    - yaml-support with type-safe .from_file() & .to_file()
        - stores minimal set (filters out unset & default parameters)
    - schema cls.schema() can also be stored to yaml with .schema_to_file()
    """

    model_config = ConfigDict(
        frozen=True,  # -> const after creation, hashable! but currently manually with .get_hash()
        extra="forbid",  # no unnamed attributes allowed
        validate_default=True,
        validate_assignment=True,  # not relevant for the frozen model
        str_min_length=1,  # force more meaningful descriptors,
        # ⤷ TODO: was 4 but localizing constraints works different with pydantic2
        #       - might be solvable with "use_enum_values=True"
        str_max_length=512,
        # ⤷ local str-length constraints overrule global ones!
        str_strip_whitespace=True,  # strip leading & trailing whitespaces
        use_enum_values=True,  # cleaner export of enum-parameters
        allow_inf_nan=False,  # float without +-inf or NaN
        # defer_build=True, possible speedup -> but it triggers a bug
    )

    def __repr__(self) -> str:
        """string-representation allows print(model)."""
        name = type(self).__name__
        content = self.model_dump(exclude_unset=True, exclude_defaults=True)
        return f"{name}({content})"

    def __str__(self) -> str:
        """string-representation allows str(model)."""
        content = yaml.safe_dump(
            self.model_dump(exclude_unset=True, exclude_defaults=True),
            default_flow_style=False,
            sort_keys=False,
        )
        return str(content)

    def __getitem__(self, key: str) -> Any:
        """Allow dict access like model["key"].

        in addition to model.key.
        """
        return self.__getattribute__(key)

    def __contains__(self, item: str) -> bool:
        """Allow checks like 'x in YClass'."""
        return item in self.model_dump().keys()  # noqa: SIM118
        # more correct, but probably slower than hasattr

    def keys(self):  # noqa: ANN201
        """Fn of dict."""
        return self.model_dump().keys()

    def items(self) -> Generator[tuple, None, None]:
        """Fn of dict."""
        for key in self.keys():
            yield key, self[key]

    @classmethod
    def schema_to_file(cls, path: str | Path) -> None:
        """Store schema to yaml (for frontend-generators)."""
        model_dict = cls.model_json_schema()
        model_yaml = yaml.safe_dump(model_dict, default_flow_style=False, sort_keys=False)
        with Path(path).resolve().with_suffix(".yaml").open("w") as f:
            f.write(model_yaml)

    def to_file(
        self,
        path: str | Path,
        comment: str | None = None,
        *,
        minimal: bool = True,
        use_pickle: bool = False,
    ) -> Path:
        """Store data to yaml in a wrapper.

        minimal: stores minimal set (filters out unset & default parameters)
        pickle: uses pickle to serialize data, on BBB >100x faster for large files
        comment: documentation.
        """
        model_wrap = Wrapper(
            datatype=type(self).__name__,
            comment=comment,
            created=local_now(),
            parameters=self.model_dump(exclude_unset=minimal),
        )
        model_dict = model_wrap.model_dump(exclude_unset=minimal, exclude_defaults=minimal)
        if use_pickle:
            model_serial = pickle.dumps(path_to_str(model_dict))
            model_path = Path(path).resolve().with_suffix(".pickle")
        else:
            # TODO: x64 windows supports CSafeLoader/dumper,
            #       there are examples that replace load if avail
            model_serial = yaml.safe_dump(
                model_dict,
                default_flow_style=False,
                sort_keys=False,
            )
            model_path = Path(path).resolve().with_suffix(".yaml")
        # TODO: handle directory

        if not model_path.parent.exists():
            model_path.parent.mkdir(parents=True)
        with model_path.open("wb" if use_pickle else "w") as f:
            f.write(model_serial)
        return model_path

    @classmethod
    def from_file(cls, path: str | Path) -> Self:
        """Load from YAML or pickle file."""
        path: Path = Path(path)
        if not Path(path).exists():
            raise FileNotFoundError
        if path.suffix.lower() == ".pickle":
            with Path(path).open("rb") as shp_file:
                shp_dict = pickle.load(shp_file)  # noqa: S301
        else:
            with Path(path).open() as shp_file:
                shp_dict = yaml.safe_load(shp_file)
        shp_wrap = Wrapper(**shp_dict)
        if shp_wrap.datatype != cls.__name__:
            raise ValueError("Model in file does not match the actual Class")
        return cls(**shp_wrap.parameters)

    def get_hash(self) -> str:
        return hashlib.sha3_224(str(self.model_dump()).encode("utf-8")).hexdigest()



"""Demonstrate behavior of Virtual Source Algorithms.

The emulation recreates an observer-cape, the virtual Source and a virtual target
- input = hdf5-file with a harvest-recording
- output = hdf5-file
- config is currently hardcoded, but it could be an emulation-task
- target is currently a simple resistor

The output file can be analyzed and plotted with shepherds tool suite.

Output:
E_out = 220.001 mWs -> direct (no current-limit)
E_out = 14.670 mWs -> diode+capacitor
E_out = 14.563 mWs -> diode+resistor+capacitor
E_out = 16.718 mWs -> BQ25504
E_out = 16.805 mWs -> BQ25504s
E_out = 16.369 mWs -> BQ25570
E_out = 16.387 mWs -> BQ25570s

"""

from pathlib import Path

from shepherd_core.data_models import VirtualSourceConfig
from shepherd_core.data_models import VirtualStorageConfig
from shepherd_core.vsource import ResistiveTarget
from shepherd_core.vsource import simulate_source

from shepherd_data import Reader

# config simulation
file_input = Path(__file__).parent / "jogging_ivcurve.h5"

src_list = [
    "direct",
    "diode+capacitor",
    "diode+resistor+capacitor",
    "BQ25504",
    "BQ25504s",
    "BQ25570",
    "BQ25570s",
]
tgt = ResistiveTarget(R_Ohm=1_000, controlled=True)
save_files = True

if not file_input.exists():
    raise FileNotFoundError(
        "Input-File not found - please run harvester-simulation first to create it."
    )

for src_name in src_list:
    file_output = file_input.with_stem(file_input.stem + "_emu_" + src_name) if save_files else None

    e_out_Ws = simulate_source(
        config=VirtualSourceConfig(
            name=src_name,
            C_output_uF=0,
            V_intermediate_enable_output_threshold_mV=1,
            V_intermediate_disable_output_threshold_mV=0,
            # jogging-dataset has maximum VOC of ~1.6 V -> lower set-point for non-boost
            storage=VirtualStorageConfig.capacitor(C_uF=100, V_rated=10.0)
            if "direct" not in src_name
            else None,
            V_pwr_good_enable_threshold_mV=1300 if "dio" in src_name else 2800,
            V_pwr_good_disable_threshold_mV=1000 if "dio" in src_name else 2400,
            V_input_drop_mV=150 if "dio" in src_name else 0,
        ),
        target=tgt,
        path_input=file_input,
        path_output=file_output,
        monitor_internals=True,
    )
    print(f"E_out = {e_out_Ws * 1e3:.3f} mWs -> {src_name}")

    with Reader(file_output, verbose=False) as reader:
        reader.plot_to_file()
        reader.plot_to_file(start_s=1, end_s=1.1)



from datetime import timedelta

import pytest
from pydantic import ValidationError
from shepherd_core.data_models import VirtualHarvesterConfig
from shepherd_core.data_models import VirtualSourceConfig
from shepherd_core.data_models.content import EnergyEnvironment
from shepherd_core.data_models.content import Firmware
from shepherd_core.data_models.experiment import Experiment
from shepherd_core.data_models.experiment import GpioActuation
from shepherd_core.data_models.experiment import GpioEvent
from shepherd_core.data_models.experiment import GpioLevel
from shepherd_core.data_models.experiment import GpioTracing
from shepherd_core.data_models.experiment import PowerTracing
from shepherd_core.data_models.experiment import SystemLogging
from shepherd_core.data_models.experiment import TargetConfig
from shepherd_core.data_models.testbed import GPIO
from shepherd_core.data_models.testbed import Target

from shepherd_core import local_now

from .conftest import load_yaml


def test_experiment_model_exp_min() -> None:
    Experiment(
        name="mex per",
        target_configs=[
            TargetConfig(
                target_IDs=[1],
                energy_env=EnergyEnvironment(name="SolarSunny"),
                firmware1=Firmware(name="nrf52_demo_rf"),
            )
        ],
    )


def test_experiment_model_exp_yaml_load() -> None:
    exp1_data = load_yaml("example_config_experiment_alternative.yaml")
    Experiment(**exp1_data)


def test_experiment_model_exp_yaml_comparison() -> None:
    exp1_data = load_yaml("example_config_experiment_alternative.yaml")
    exp1 = Experiment(**exp1_data)

    target_cfgs = TargetConfig(
        target_IDs=list(range(1, 5)),
        custom_IDs=list(range(4)),
        energy_env={"name": "SolarSunny"},
        virtual_source={"name": "diode+capacitor"},
        firmware1={"name": "nrf52_demo_rf"},
    )
    exp2 = Experiment(
        name="meaningful Test-Name",
        time_start="2033-12-12 12:12:12",
        target_configs=[target_cfgs],
    )
    assert exp1.model_dump() == exp2.model_dump()
    assert exp1.get_hash() == exp2.get_hash()


def test_experiment_model_exp_collision_target_id() -> None:
    hrv = VirtualHarvesterConfig(name="mppt_bq_thermoelectric")
    target_cfgs = [
        TargetConfig(
            target_IDs=[1, 2, 3],  # <- collision
            custom_IDs=[0, 1, 2, 3],
            energy_env={"name": "SolarSunny"},
            virtual_source={"name": "diode+capacitor"},
            firmware1={"name": "nrf52_demo_rf"},
        ),
        TargetConfig(
            target_IDs=list(range(1, 5)),  # <- collision
            custom_IDs=list(range(7, 18)),  # note: longer list is OK
            energy_env=EnergyEnvironment(name="ThermoelectricWashingMachine"),
            virtual_source=VirtualSourceConfig(name="BQ25570-Schmitt", harvester=hrv),
            firmware1=Firmware(name="nrf52_demo_rf"),
            firmware2=Firmware(name="msp430_deep_sleep"),
        ),
    ]
    with pytest.raises(ValidationError):
        _ = Experiment(
            name="meaningful Test-Name",
            time_start=local_now() + timedelta(minutes=30),
            target_configs=target_cfgs,
        )


def test_experiment_model_exp_collision_custom_id() -> None:
    hrv = VirtualHarvesterConfig(name="mppt_bq_thermoelectric")
    target_cfgs = [
        TargetConfig(
            target_IDs=[1, 2, 3],
            custom_IDs=[0, 1, 7],  # <- collision
            energy_env={"name": "SolarSunny"},
            virtual_source={"name": "diode+capacitor"},
            firmware1={"name": "nrf52_demo_rf"},
        ),
        TargetConfig(
            target_IDs=list(range(1, 5)),
            custom_IDs=list(range(7, 18)),  # <- collision
            energy_env=EnergyEnvironment(name="ThermoelectricWashingMachine"),
            virtual_source=VirtualSourceConfig(name="BQ25570-Schmitt", harvester=hrv),
            firmware1=Firmware(name="nrf52_demo_rf"),
            firmware2=Firmware(name="msp430_deep_sleep"),
        ),
    ]
    with pytest.raises(ValidationError):
        _ = Experiment(
            name="meaningful Test-Name",
            time_start=local_now() + timedelta(minutes=30),
            target_configs=target_cfgs,
        )


def test_experiment_model_exp_collision_observer() -> None:
    hrv = VirtualHarvesterConfig(name="mppt_bq_thermoelectric")
    target_cfgs = [
        TargetConfig(
            target_IDs=[1, 1001],  # <- both on same observer
            custom_IDs=list(range(7, 18)),
            energy_env=EnergyEnvironment(name="ThermoelectricWashingMachine"),
            virtual_source=VirtualSourceConfig(name="BQ25570-Schmitt", harvester=hrv),
            firmware1=Firmware(name="nrf52_demo_rf"),
            firmware2=Firmware(name="msp430_deep_sleep"),
        ),
    ]
    with pytest.raises(ValidationError):
        _ = Experiment(
            name="meaningful Test-Name",
            time_start=local_now() + timedelta(minutes=30),
            target_configs=target_cfgs,
        )


def test_experiment_model_exp_missing_target() -> None:
    hrv = VirtualHarvesterConfig(name="mppt_bq_thermoelectric")
    with pytest.raises(ValidationError):
        # should raise ValueError in Experiment
        # buts gets already caught in target_config
        _ = [
            TargetConfig(
                target_IDs=[1234567],  # <- not existent
                custom_IDs=list(range(7, 18)),
                energy_env=EnergyEnvironment(name="ThermoelectricWashingMachine"),
                virtual_source=VirtualSourceConfig(name="BQ25570-Schmitt", harvester=hrv),
                firmware1=Firmware(name="nrf52_demo_rf"),
                firmware2=Firmware(name="msp430_deep_sleep"),
            ),
        ]
        # experiment (like above) removed


def test_experiment_model_pwrtracing_min() -> None:
    PowerTracing()


def test_experiment_model_pwrtracing_power_only() -> None:
    PowerTracing(only_power=True)


def test_experiment_model_pwrtracing_sample_rate() -> None:
    PowerTracing(samplerate=10)
    PowerTracing(samplerate=100)
    PowerTracing(samplerate=100000)


def test_experiment_model_pwrtracing_sample_rate_fail() -> None:
    with pytest.raises(ValidationError):
        PowerTracing(samplerate=1)
    with pytest.raises(ValidationError):
        PowerTracing(samplerate=111)


def test_experiment_model_gpiotracing_min() -> None:
    GpioTracing()


def test_experiment_model_gpiotracing_fault_mask() -> None:
    with pytest.raises(ValidationError):
        GpioTracing(mask=0)


def test_experiment_model_gpioevent_min() -> None:
    gevt = GpioEvent(
        delay=300,
        gpio=GPIO(name="GPIO3"),
        level=GpioLevel.high,
    )
    gevt.get_events()


def test_experiment_model_gpioevent_fault_readonly() -> None:
    with pytest.raises(ValidationError):
        GpioEvent(
            delay=300,
            gpio=GPIO(name="BAT_OK"),
            level=GpioLevel.high,
        )


@pytest.mark.skip(reason="Not implemented ATM")
def test_experiment_model_gpioactuation_min() -> None:
    gact = GpioActuation(
        events=[
            GpioEvent(
                delay=300,
                gpio=GPIO(name="GPIO2"),
                level=GpioLevel.high,
            )
        ]
    )
    gact.get_gpios()


def test_experiment_model_syslogging_min() -> None:
    SystemLogging()


def test_experiment_model_tgt_cfg_min1() -> None:
    cfg = TargetConfig(
        target_IDs=[1],
        energy_env=EnergyEnvironment(name="SolarSunny"),
        firmware1=Firmware(name="nrf52_demo_rf"),
    )
    for _id in cfg.target_IDs:
        Target(id=_id)
    assert cfg.get_custom_id(1) is None


def test_experiment_model_tgt_cfg_min2() -> None:
    cfg = TargetConfig(
        target_IDs=[1, 2],
        custom_IDs=[7, 9],
        energy_env=EnergyEnvironment(name="SolarSunny"),
        firmware1=Firmware(name="nrf52_demo_rf"),
    )
    for _id in cfg.target_IDs:
        Target(id=_id)
    assert cfg.get_custom_id(1) == 7
    assert cfg.get_custom_id(2) == 9
    assert cfg.get_custom_id(3) is None


def test_experiment_model_tgt_cfg_fault_valid_ee() -> None:
    with pytest.raises(ValidationError):
        _ = TargetConfig(
            target_IDs=[1],
            energy_env=EnergyEnvironment(name="nuclear"),
            firmware1=Firmware(name="nrf52_demo_rf"),
        )


def test_experiment_model_tgt_cfg_fault_firmware1() -> None:
    with pytest.raises(ValidationError):
        _ = TargetConfig(
            target_IDs=[1],  # is nRF
            energy_env=EnergyEnvironment(name="SolarSunny"),
            firmware1=Firmware(name="msp430_spi_fram"),
        )


def test_experiment_model_tgt_cfg_fault_firmware2() -> None:
    with pytest.raises(ValidationError):
        _ = TargetConfig(
            target_IDs=[1],  # is nRF
            energy_env=EnergyEnvironment(name="SolarSunny"),
            firmware1=Firmware(name="nrf52_demo_rf"),
            firmware2=Firmware(name="nrf52_demo_rf"),
        )


def test_experiment_model_tgt_cfg_fault_custom_ids() -> None:
    with pytest.raises(ValidationError):
        _ = TargetConfig(
            target_IDs=[1, 2, 3],
            custom_IDs=[0, 1],  # not enough
            energy_env=EnergyEnvironment(name="SolarSunny"),
            firmware1=Firmware(name="nrf52_demo_rf"),
        )



"""Simulate behavior of virtual source algorithms.

The simulation recreates an observer-cape, the virtual Source and a virtual target
- input = hdf5-file with a harvest-recording
- output = optional as hdf5-file

The output file can be analyzed and plotted with shepherds tool suite.
"""

from contextlib import ExitStack
from pathlib import Path

import numpy as np
from tqdm import tqdm

from shepherd_core.data_models.base.calibration import CalibrationEmulator
from shepherd_core.data_models.content.virtual_source_config import VirtualSourceConfig
from shepherd_core.logger import log
from shepherd_core.reader import Reader
from shepherd_core.writer import Writer

from .target_model import TargetABC
from .virtual_source_model import VirtualSourceModel


def simulate_source(
    config: VirtualSourceConfig,
    target: TargetABC,
    path_input: Path,
    path_output: Path | None = None,
    *,
    monitor_internals: bool = False,
) -> float:
    """Simulate behavior of virtual source algorithms.

    FN returns the consumed energy of the target.
    """
    with ExitStack() as stack:
        file_inp = Reader(path_input, verbose=False)
        stack.enter_context(file_inp)
        cal_emu = CalibrationEmulator()
        cal_inp = file_inp.get_calibration_data()

        if path_output:
            file_out = Writer(
                path_output, cal_data=cal_emu, mode="emulator", verbose=False, force_overwrite=True
            )
            stack.enter_context(file_out)
            file_out.store_hostname("emu_sim_" + config.name)
            file_out.store_config(config.model_dump())
            cal_out = file_out.get_calibration_data()

        src = VirtualSourceModel(
            config,
            cal_emu,
            dtype_in=file_inp.get_datatype(),
            log_intermediate=False,
            window_size=file_inp.get_window_samples(),
            voltage_step_V=file_inp.get_voltage_step(),
        )
        i_out_nA = 0
        e_out_Ws = 0.0
        if monitor_internals and path_output:
            stats_sample = 0
            stats_internal = np.empty((round(file_inp.runtime_s * file_inp.samplerate_sps), 11))
            try:
                # keep dependencies low
                from matplotlib import pyplot as plt  # noqa: PLC0415
            except ImportError:
                log.warning("Matplotlib not installed, plotting of internals disabled")
                stats_internal = None
        else:
            stats_internal = None

        for _t, v_inp, i_inp in tqdm(
            file_inp.read(is_raw=True), total=file_inp.chunks_n, desc="Chunk", leave=False
        ):
            v_uV = 1e6 * cal_inp.voltage.raw_to_si(v_inp)
            i_nA = 1e9 * cal_inp.current.raw_to_si(i_inp)

            for _n in range(len(_t)):
                v_uV[_n] = src.iterate_sampling(
                    V_inp_uV=int(v_uV[_n]),
                    I_inp_nA=int(i_nA[_n]),
                    I_out_nA=i_out_nA,
                )
                i_out_nA = target.step(int(v_uV[_n]), pwr_good=src.cnv.get_power_good())
                i_nA[_n] = i_out_nA

                if stats_internal is not None:
                    stats_internal[stats_sample] = [
                        _t[_n] * 1e-9,  # s
                        src.hrv.voltage_hold * 1e-6,
                        src.cnv.V_input_request_uV * 1e-6,  # V
                        src.hrv.voltage_set_uV * 1e-6,
                        src.cnv.V_mid_uV * 1e-6,
                        src.hrv.current_hold * 1e-6,  # mA
                        src.hrv.current_delta * 1e-6,
                        i_out_nA * 1e-6,
                        src.cnv.P_inp_fW * 1e-12,  # mW
                        src.cnv.P_out_fW * 1e-12,
                        src.cnv.get_power_good(),
                    ]
                    stats_sample += 1

            e_out_Ws += (v_uV * i_nA).sum() * 1e-15 * file_inp.sample_interval_s
            if path_output:
                v_out = cal_out.voltage.si_to_raw(1e-6 * v_uV)
                i_out = cal_out.current.si_to_raw(1e-9 * i_nA)
                file_out.append_iv_data_raw(_t, v_out, i_out)

    if stats_internal is not None:
        stats_internal = stats_internal[:stats_sample, :]
        fig, axs = plt.subplots(4, 1, sharex="all", figsize=(20, 4 * 6), layout="tight")
        fig.suptitle(f"VSrc-Sim with {config.name}, Inp={path_input.name}, E={e_out_Ws} Ws")
        axs[0].set_ylabel("Voltages [V]")
        axs[0].plot(stats_internal[:, 0], stats_internal[:, 1:5])
        axs[0].legend(["V_cv_hold", "V_inp_Req", "V_cv_set", "V_cap"], loc="upper right")

        axs[1].set_ylabel("Current [mA]")
        axs[1].plot(stats_internal[:, 0], stats_internal[:, 5:8])
        axs[1].legend(["C_cv_hold", "C_cv_delta", "C_out"], loc="upper right")

        axs[2].set_ylabel("Power [mW]")
        axs[2].plot(stats_internal[:, 0:1], stats_internal[:, 8:10])
        axs[2].legend(["P_inp", "P_out"], loc="upper right")

        axs[3].set_ylabel("PwrGood [n]")
        axs[3].plot(stats_internal[:, 0], stats_internal[:, 10])
        axs[3].legend(["PwrGood"], loc="upper right")

        axs[3].set_xlabel("Runtime [s]")

        for ax in axs:
            # deactivates offset-creation for ax-ticks
            ax.get_yaxis().get_major_formatter().set_useOffset(False)
            ax.get_xaxis().get_major_formatter().set_useOffset(False)

        plt.savefig(path_output.with_suffix(".png"))
        plt.close(fig)
        plt.clf()

    return e_out_Ws



"""Container for a common configuration.

This can be adapted by the user by importing 'config' and changing its variables.
"""

from pydantic import BaseModel
from pydantic import HttpUrl


class ConfigDefault(BaseModel):
    """Container for a common configuration."""

    __slots__ = ()

    TESTBED: str = "shepherd_tud_nes"
    """name of the testbed to validate against - if enabled - see switch below"""
    VALIDATE_INFRA: bool = True
    """switch to turn on / off deep validation of data models also considering the current
    layout & infrastructure of the testbed.
    """

    SAMPLERATE_SPS: int = 100_000
    """Rate of IV-Recording of the testbed."""

    UID_NAME: str = "SHEPHERD_NODE_ID"
    """Variable name to patch in ELF-file"""
    UID_SIZE: int = 2
    """Variable size in Byte"""

    TESTBED_SERVER: HttpUrl = "https://shepherd.cfaed.tu-dresden.de:8000/"
    """Server that holds up to date testbed fixtures"""


config = ConfigDefault()



"""Virtual targets with different characteristics.

TODO: add more targets
  - diode
  - constant power
  - constant current
  - msp430 (const I)
  - nRF (constant power due to regulator)
  - riotee
"""

import math
from abc import ABC
from abc import abstractmethod
from contextlib import suppress


class TargetABC(ABC):
    """Abstract base class for all targets."""

    @abstractmethod
    def step(self, voltage_uV: int, *, pwr_good: bool) -> float:
        """Calculate one time step and return drawn current in nA."""


class ResistiveTarget(TargetABC):
    """Predictable target with linear behavior."""

    def __init__(self, R_Ohm: float, *, controlled: bool = False) -> None:
        if R_Ohm <= 1e-3:
            raise ValueError("Resistance must be greater than 1 mOhm.")
        self.R_kOhm = 1e-3 * R_Ohm
        self.ctrl = controlled

    def step(self, voltage_uV: int, *, pwr_good: bool) -> float:
        if pwr_good or not self.ctrl:
            return voltage_uV / self.R_kOhm  # = nA
        return 0


class DiodeTarget(TargetABC):
    """Emulate a diode and current limiting resistor in series.

    Good for modeling a debug-diode that burns energy.
    It uses Shockley Diode Equation to estimate a model for diode
    I_D = I_S * ( e ^ ( U_D / n*U_T ) - 1 )

    diode of shepherd target:
    d1 = DiodeTarget(V_forward_V = 2.0, I_forward_A = 20e-3, R_Ohm = 100)

    """

    def __init__(
        self,
        V_forward_V: float,
        I_forward_A: float,
        R_Ohm: float,
        *,
        controlled: bool = False,
    ) -> None:
        if R_Ohm <= 1e-3:
            raise ValueError("Resistance must be greater than 1 mOhm.")
        if V_forward_V <= 0.2:
            raise ValueError("Forward-Voltage of diode must be greater than 200 mV.")
        if I_forward_A <= 0:
            raise ValueError("Forward-current of diode must be greater than 0 A.")

        k = 1.380649e-23  # boltzmann
        q = 1.6021766e-19  # elementary charge
        TJ = 100 + 273.15  # junction temperature

        V_T = k * TJ / q  # thermal voltage
        n = 2  # ideality factor
        self.c1 = V_T * n
        # NOTE: math.expm1(x) = math.exp(x) - 1 = e^x -1
        self.I_S = I_forward_A / math.expm1(V_forward_V / self.c1)  # scale current
        self.R_Ohm = R_Ohm
        self.ctrl = controlled

    def step(self, voltage_uV: int, *, pwr_good: bool) -> float:
        if pwr_good or not self.ctrl:
            V_CC: float = voltage_uV * 1e-6
            V_D: float = V_CC / 2
            I_R = I_D = 0.0
            # there is no direct formular, but this iteration converges fast
            for _ in range(10):
                # low voltages tend to produce log(x<0)=err
                with suppress(ValueError):
                    V_D = self.c1 * math.log(1 + (V_CC - V_D) / (self.R_Ohm * self.I_S))
                # both currents are positive and should be identical
                I_R = max(0.0, (V_CC - V_D) / self.R_Ohm)
                I_D = max(0.0, self.I_S * math.expm1(V_D / self.c1))
                with suppress(ZeroDivisionError):
                    if abs(I_R / I_D - 1) < 1e-6:
                        break
                # take mean of both currents and determine a new V_D
                V_D = V_CC - self.R_Ohm * (I_R + I_D) / 2
            return 1e9 * (I_R + I_D) / 2  # = nA
        return 0


class ConstantCurrentTarget(TargetABC):
    """Recreate simple MCU without integrated regulator, i.e. msp430."""

    def __init__(self, I_active_A: float, I_sleep_A: float = 0) -> None:
        if I_active_A <= 0 or I_sleep_A < 0:
            raise ValueError("Current must be greater than 0.")
        self.I_active_nA = 1e9 * I_active_A
        self.I_sleep_nA = 1e9 * I_sleep_A

    def step(self, voltage_uV: int, *, pwr_good: bool) -> float:  # noqa: ARG002
        return self.I_active_nA if pwr_good else self.I_sleep_nA


class ConstantPowerTarget(TargetABC):
    """Recreate MCU with integrated regulator, i.e. nRF52."""

    def __init__(self, P_active_W: float, P_sleep_W: float = 0) -> None:
        if P_active_W <= 0 or P_sleep_W < 0:
            raise ValueError("Power must be greater than 0.")
        self.P_active_fW = 1e15 * P_active_W
        self.P_sleep_fW = 1e15 * P_sleep_W

    def step(self, voltage_uV: int, *, pwr_good: bool) -> float:
        return (self.P_active_fW if pwr_good else self.P_sleep_fW) / voltage_uV  # = nA


# exemplary instantiations

diode_target_burn = DiodeTarget(V_forward_V=2.0, I_forward_A=20e-3, R_Ohm=100)
mcu_msp430fr = ConstantCurrentTarget(I_active_A=16 * 118e-6, I_sleep_A=350e-9)
mcu_msp_deep_sleep = ConstantCurrentTarget(45e-9, 45e-9)
# TODO: writing FRAM
mcu_nrf52840 = ConstantPowerTarget(P_active_W=3 * 3e-3, P_sleep_W=3 * 0.97e-6)
mcu_nrf_tx_8dBm = ConstantPowerTarget(P_active_W=3.0 * 16.40e-3)
mcu_nrf_tx_0dBm = ConstantPowerTarget(P_active_W=3.0 * 6.40e-3)
mcu_nrf_rx = ConstantPowerTarget(P_active_W=3.0 * 6.26e-3)
# data based on PS1.10
# - TX 8 dBm -> 6.31 mW -> DS shows 16.40 mA @ 3V (DC/DC)
# - TX 0 dBm -> 1 mW ->              6.40 mA
# - RX ->                            6.26 mA



"""Container for commonly shared constants."""

from .config import config

# TODO: deprecated - replace with config in subprojects and then remove here
SAMPLERATE_SPS_DEFAULT: int = config.SAMPLERATE_SPS



"""Writer that inherits from Reader-Baseclass."""

import logging
import math
import pathlib
from collections.abc import Mapping
from datetime import timedelta
from itertools import product
from pathlib import Path
from types import TracebackType
from typing import Any

import h5py
import numpy as np
import yaml
from pydantic import validate_call
from typing_extensions import Self
from yaml import Node
from yaml import SafeDumper

from .config import config
from .data_models.base.calibration import CalibrationEmulator as CalEmu
from .data_models.base.calibration import CalibrationHarvester as CalHrv
from .data_models.base.calibration import CalibrationSeries as CalSeries
from .data_models.content.energy_environment import EnergyDType
from .data_models.task import Compression
from .data_models.task.emulation import c_translate
from .reader import Reader


# copy of core/models/base/shepherd - needed also here
def path2str(
    dumper: SafeDumper, data: pathlib.Path | pathlib.WindowsPath | pathlib.PosixPath
) -> Node:
    """Add a yaml-representation for a specific datatype."""
    return dumper.represent_scalar("tag:yaml.org,2002:str", str(data.as_posix()))


def time2int(dumper: SafeDumper, data: timedelta) -> Node:
    """Add a yaml-representation for a specific datatype."""
    return dumper.represent_scalar("tag:yaml.org,2002:int", str(int(data.total_seconds())))


yaml.add_representer(pathlib.PosixPath, path2str, SafeDumper)
yaml.add_representer(pathlib.WindowsPath, path2str, SafeDumper)
yaml.add_representer(pathlib.Path, path2str, SafeDumper)
yaml.add_representer(timedelta, time2int, SafeDumper)


def unique_path(base_path: str | Path, suffix: str) -> Path:
    """Find an unused filename in case it already exists.

    :param base_path: file-path to test
    :param suffix: file-suffix
    :return: new non-existing path
    """
    counter = 0
    while True:
        path = Path(base_path).with_suffix(f".{counter}{suffix}")
        if not path.exists():
            return path
        counter += 1


class Writer(Reader):
    """Stores data for Shepherd in HDF5 format.

    Choose lossless compression filter
     - lzf:  low to moderate compression, VERY fast, no options
             -> 20 % cpu overhead for half the filesize
     - gzip: good compression, moderate speed, select level from 1-9, default is 4
             -> lower levels seem fine
             -> _algo=number instead of "gzip" is read as compression level for gzip
     -> comparison / benchmarks https://www.h5py.org/lzf/

    Args:
    ----
        file_path: (Path) Name of the HDF5 file that data will be written to
        mode: (str) Indicates if this is data from harvester or emulator
        datatype: (str) choose type: ivsample (most common), ivcurve or isc_voc
        window_samples: (int) windows size for the datatype ivcurve
        cal_data: (CalibrationData) Data is written as raw ADC
            values. We need calibration data in order to convert to physical
            units later.
        modify_existing: (bool) explicitly enable modifying existing file
            otherwise a unique name will be found
        compression: (str) use either None, lzf or "1" (gzips compression level)
        verbose: (bool) provides more debug-info

    """

    MODE_DEFAULT: str = "harvester"
    DATATYPE_DEFAULT: EnergyDType = EnergyDType.ivsample

    _CHUNK_SHAPE: tuple = (Reader.CHUNK_SAMPLES_N,)

    @validate_call
    def __init__(
        self,
        file_path: Path,
        mode: str | None = None,
        datatype: str | EnergyDType | None = None,
        window_samples: int | None = None,
        cal_data: CalSeries | CalEmu | CalHrv | None = None,
        compression: Compression | None = Compression.default,
        *,
        modify_existing: bool = False,
        force_overwrite: bool = False,
        verbose: bool = True,
    ) -> None:
        self._modify = modify_existing
        if compression is not None:
            self._compression = c_translate[compression.value]
        else:
            self._compression = None

        if not hasattr(self, "_logger"):
            self._logger: logging.Logger = logging.getLogger("SHPCore.Writer")
        # -> logger gets configured in reader()

        if self._modify or force_overwrite or not file_path.exists():
            file_path = file_path.resolve()
            self._logger.info("Storing data to   '%s'", file_path)
        elif file_path.exists() and not file_path.is_file():
            msg = f"Path is not a file ({file_path})"
            raise TypeError(msg)
        else:
            base_dir = file_path.resolve().parents[0]
            file_path_new = unique_path(base_dir / file_path.stem, file_path.suffix)
            self._logger.warning(
                "File '%s' already exists -> storing under '%s' instead",
                file_path,
                file_path_new.name,
            )
            file_path = file_path_new

        # open file
        if self._modify:
            self.h5file = h5py.File(file_path, "r+")  # = rw
        else:
            if not file_path.parent.exists():
                file_path.parent.mkdir(parents=True)
            self.h5file = h5py.File(file_path, "w")
            # ⤷ write, truncate if exist
            self._create_skeleton()

        # Handle Mode
        if isinstance(mode, str) and mode not in self.MODE_TO_DTYPE:
            msg = f"Can't handle mode '{mode}' (choose one of {self.MODE_TO_DTYPE})"
            raise ValueError(msg)

        if mode is not None:
            self.h5file.attrs["mode"] = mode
        if "mode" not in self.h5file.attrs:
            self.h5file.attrs["mode"] = self.MODE_DEFAULT

        _dtypes = self.MODE_TO_DTYPE[self.get_mode()]

        # Handle Datatype
        if isinstance(datatype, str):
            datatype = EnergyDType[datatype]
        if isinstance(datatype, EnergyDType) and datatype not in _dtypes:
            msg = f"Can't handle value '{datatype}' of datatype (choose one of {_dtypes})"
            raise ValueError(msg)

        if isinstance(datatype, EnergyDType):
            self.h5file["data"].attrs["datatype"] = datatype.name
        if "datatype" not in self.h5file["data"].attrs:
            self.h5file["data"].attrs["datatype"] = self.DATATYPE_DEFAULT.name
        if self.get_datatype() not in _dtypes:
            msg = (
                f"Can't handle value '{self.get_datatype()}' of datatype (choose one of {_dtypes})"
            )
            raise ValueError(msg)

        # Handle Window_samples
        if window_samples is not None:
            self.h5file["data"].attrs["window_samples"] = window_samples
        if "window_samples" not in self.h5file["data"].attrs:
            self.h5file["data"].attrs["window_samples"] = 0

        if datatype == EnergyDType.ivcurve and self.get_window_samples() < 1:
            raise ValueError("Window Size argument needed for ivcurve-Datatype")

        # Handle Cal
        if isinstance(cal_data, (CalEmu, CalHrv)):
            cal_data = CalSeries.from_cal(cal_data)

        if isinstance(cal_data, CalSeries):
            for ds, param in product(["current", "voltage", "time"], ["gain", "offset"]):
                self.h5file["data"][ds].attrs[param] = cal_data[ds][param]
        else:
            # check if there are unset cal-values and set them to default
            cal_data = CalSeries()
            for ds, param in product(["current", "voltage", "time"], ["gain", "offset"]):
                if param not in self.h5file["data"][ds].attrs:
                    self.h5file["data"][ds].attrs[param] = cal_data[ds][param]

        # show key parameters for h5-performance
        settings = list(self.h5file.id.get_access_plist().get_cache())
        self._logger.debug("H5Py Cache_setting=%s (_mdc, _nslots, _nbytes, _w0)", settings)

        super().__init__(file_path=file_path, verbose=verbose)

    def __enter__(self) -> Self:
        super().__enter__()
        return self

    def __exit__(
        self,
        typ: type[BaseException] | None = None,
        exc: BaseException | None = None,
        tb: TracebackType | None = None,
        extra_arg: int = 0,
    ) -> None:
        self._align()
        self._refresh_file_stats()
        self._logger.info(
            "closing hdf5 file, %.1f s iv-data, size = %.3f MiB, rate = %.0f KiB/s",
            self.runtime_s,
            self.file_size / 2**20,
            self.data_rate / 2**10,
        )
        self.is_valid()
        self.h5file.close()

    def _create_skeleton(self) -> None:
        """Initialize the structure of the HDF5 file.

        HDF5 is hierarchically structured and before writing data, we have to
        set up this structure, i.e. creating the right groups with corresponding
        data types. We will store 3 types of data in a database: The
        actual IV samples recorded either from the harvester (during recording)
        or the target (during emulation). Any log messages, that can be used to
        store relevant events or tag some parts of the recorded data.

        """
        # Store voltage and current samples in the data group,
        # both are stored as 4 Byte unsigned int
        grp_data = self.h5file.create_group("data")
        # the size of window_samples-attribute in harvest-data indicates ivsurface / curves as input
        # -> emulator uses virtual-harvester, field will be adjusted by .embed_config()
        grp_data.attrs["window_samples"] = 0

        grp_data.create_dataset(
            "time",
            (0,),
            dtype="u8",
            maxshape=(None,),
            chunks=self._CHUNK_SHAPE,
            compression=self._compression,
        )
        grp_data["time"].attrs["unit"] = "s"
        grp_data["time"].attrs["description"] = "system time [s] = value * gain + (offset)"

        grp_data.create_dataset(
            "current",
            (0,),
            dtype="u4",
            maxshape=(None,),
            chunks=self._CHUNK_SHAPE,
            compression=self._compression,
        )
        grp_data["current"].attrs["unit"] = "A"
        grp_data["current"].attrs["description"] = "current [A] = value * gain + offset"

        grp_data.create_dataset(
            "voltage",
            (0,),
            dtype="u4",
            maxshape=(None,),
            chunks=self._CHUNK_SHAPE,
            compression=self._compression,
        )
        grp_data["voltage"].attrs["unit"] = "V"
        grp_data["voltage"].attrs["description"] = "voltage [V] = value * gain + offset"

    def append_iv_data_raw(
        self,
        timestamp: np.ndarray | float | int,  # noqa: PYI041
        voltage: np.ndarray,
        current: np.ndarray,
    ) -> None:
        """Write raw data to database.

        Args:
        ----
            timestamp: just start of chunk (1 timestamp) or whole ndarray
            voltage: ndarray as raw unsigned integers
            current: ndarray as raw unsigned integers

        """
        len_new = min(voltage.size, current.size)

        if isinstance(timestamp, float):
            timestamp = int(timestamp)
        if isinstance(timestamp, int):
            time_series_ns = self.sample_interval_ns * np.arange(len_new).astype("u8")
            timestamp = timestamp + time_series_ns
        if isinstance(timestamp, np.ndarray):
            len_new = min(len_new, timestamp.size)
        else:
            raise TypeError("timestamp-data was not usable")

        len_old = self.ds_voltage.shape[0]

        # resize dataset
        self.ds_time.resize((len_old + len_new,))
        self.ds_voltage.resize((len_old + len_new,))
        self.ds_current.resize((len_old + len_new,))

        # append new data
        self.ds_time[len_old : len_old + len_new] = timestamp[:len_new]
        self.ds_voltage[len_old : len_old + len_new] = voltage[:len_new]
        self.ds_current[len_old : len_old + len_new] = current[:len_new]

    def append_iv_data_si(
        self,
        timestamp: np.ndarray | float,
        voltage: np.ndarray,
        current: np.ndarray,
    ) -> None:
        """Write data (provided in SI / physical unit) to file.

        This will convert it to raw-data first.

           SI-value [SI-Unit] = raw-value * gain + offset,

        Args:
        ----
            timestamp: python timestamp (time.time()) in seconds (si-unit)
                       -> provide start of chunk (1 timestamp) or whole ndarray
            voltage: ndarray in physical-unit V
            current: ndarray in physical-unit A

        """
        # TODO: make timestamp optional to add it raw, OR unify append with granular raw-switch
        timestamp = self._cal.time.si_to_raw(timestamp)
        voltage = self._cal.voltage.si_to_raw(voltage)
        current = self._cal.current.si_to_raw(current)
        self.append_iv_data_raw(timestamp, voltage, current)

    def _align(self) -> None:
        """Align datasets with chunk-size of shepherd."""
        self._refresh_file_stats()
        chunks_n = self.ds_voltage.size / self.CHUNK_SAMPLES_N
        size_new = int(math.floor(chunks_n) * self.CHUNK_SAMPLES_N)
        if size_new < self.ds_voltage.size:
            if self.samplerate_sps != config.SAMPLERATE_SPS:
                self._logger.debug("skipped alignment due to altered samplerate")
                return
            self._logger.info(
                "aligning with chunk-size, discarding last %d entries",
                self.ds_voltage.size - size_new,
            )
            self.ds_time.resize((size_new,))
            self.ds_voltage.resize((size_new,))
            self.ds_current.resize((size_new,))

    def __setitem__(self, key: str, item: Any) -> None:
        """Conveniently store relevant key-value data (attribute) in H5-structure."""
        self.h5file.attrs.__setitem__(key, item)

    def store_config(self, data: Mapping) -> None:
        """Get a better self-describing Output-File.

        TODO: use data-model?
        :param data: from virtual harvester or converter / source.
        """
        self.h5file.attrs["config"] = yaml.safe_dump(
            data, default_flow_style=False, sort_keys=False
        )

    def store_hostname(self, name: str) -> None:
        """Option to distinguish the host, target or data-source -> perfect for plotting later.

        :param name: something unique, or "artificial" in case of generated content
        """
        self.h5file.attrs["hostname"] = name



"""Creates an overview for shepherd-host-machines.

This will collect:
- relevant software-versions
- system-parameters
- hardware-config.
"""

from datetime import datetime
from datetime import timedelta
from pathlib import Path
from typing import Annotated

from pydantic import Field
from typing_extensions import Self

from shepherd_core.data_models import ShpModel

from .python import PythonInventory
from .system import SystemInventory
from .target import TargetInventory

__all__ = [
    "Inventory",
    "InventoryList",
    "PythonInventory",
    "SystemInventory",
    "TargetInventory",
]


class Inventory(PythonInventory, SystemInventory, TargetInventory):
    """Complete inventory for one device.

    Has all child-parameters.
    """

    hostname: str
    created: datetime

    @classmethod
    def collect(cls) -> Self:
        # one by one for more precise error messages
        # NOTE: system is first, as it must take a precise timestamp
        sid = SystemInventory.collect().model_dump(exclude_unset=True, exclude_defaults=True)
        pid = PythonInventory.collect().model_dump(exclude_unset=True, exclude_defaults=True)
        tid = TargetInventory.collect().model_dump(exclude_unset=True, exclude_defaults=True)
        model = {**pid, **sid, **tid}
        # make important metadata available at root level
        model["created"] = sid["timestamp"]
        model["hostname"] = sid["hostname"]
        return cls(**model)


class InventoryList(ShpModel):
    """Collection of inventories for several devices."""

    elements: Annotated[list[Inventory], Field(min_length=1)]

    def to_csv(self, path: Path) -> None:
        """Generate a CSV.

        TODO: pretty messed up (raw lists and dicts for sub-elements)
        numpy.savetxt -> too basic
        np.concatenate(content).reshape((len(content), len(content[0]))).
        """
        if path.is_dir():
            path = path / "inventory.yaml"
        with path.resolve().open("w") as fd:
            fd.write(", ".join(self.elements[0].model_dump().keys()) + "\r\n")
            for item in self.elements:
                content = list(item.model_dump().values())
                content = ["" if value is None else str(value) for value in content]
                fd.write(", ".join(content) + "\r\n")

    def warn(self) -> dict:
        warnings = {}
        ts_earl = min([_e.created.timestamp() for _e in self.elements])
        for _e in self.elements:
            if _e.uptime > timedelta(hours=30).total_seconds():
                warnings["uptime"] = f"[{self.hostname}] restart is recommended"
            if (_e.created.timestamp() - ts_earl) > 10:
                warnings["time_delta"] = f"[{self.hostname}] time-sync has failed"

        # turn  dict[hostname][type] = val
        # to    dict[type][val] = list[hostnames]
        _inp = {
            _e.hostname: _e.model_dump(exclude_unset=True, exclude_defaults=True)
            for _e in self.elements
        }
        result = {}
        for _host, _types in _inp.items():
            for _type, _val in _types.items():
                if _type not in result:
                    result[_type] = {}
                if _val not in result[_type]:
                    result[_type][_val] = []
                result[_type][_val].append(_host)
        rescnt = {_key: len(_val) for _key, _val in result.items()}
        t_unique = [
            "h5py",
            "numpy",
            "pydantic",
            "python",
            "shepherd_core",
            "shepherd_sheep",
            "yaml",
            "zstandard",
        ]
        for _key in t_unique:
            if rescnt[_key] > 1:
                warnings[_key] = f"[{_key}] VersionMismatch - {result[_key]}"

        # TODO: finish with more potential warnings
        return warnings



"""Configs for observer features like gpio- & power-tracing."""

from datetime import timedelta
from enum import Enum
from typing import Annotated

import numpy as np
from annotated_types import Interval
from pydantic import Field
from pydantic import PositiveFloat
from pydantic import model_validator
from typing_extensions import Self

from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.testbed.gpio import GPIO
from shepherd_core.logger import log

# defaults (pre-init complex types)
zero_duration = timedelta(seconds=0)


class PowerTracing(ShpModel, title="Config for Power-Tracing"):
    """Configuration for recording the Power-Consumption of the Target Nodes."""

    intermediate_voltage: bool = False
    """
    ⤷ for EMU: record storage capacitor instead of output (good for V_out = const)
               this also includes current!
    """
    # time
    delay: timedelta = zero_duration
    """start recording after experiment started"""
    duration: timedelta | None = None  # till EOF
    """duration of recording after delay starts the process.

    default is None, recording till EOF"""

    # further processing of IV-Samples
    only_power: bool = False
    """ ⤷ reduce file-size by calculating power and automatically discard I&V
        Caution: increases cpu-utilization on observer - power @ 100 kHz is not recommended
    """
    samplerate: Annotated[int, Field(ge=10, le=100_000)] = 100_000
    """ ⤷ reduce file-size by re-sampling (mean over x samples)

        Caution: this option is available for IV-Samples (not only .only_power)
                 but results might be faulty for use-cases that are not const-voltage
                 Sum(U * I) != Sum(U) * Sum(I)
        Timestamps will be taken from the start of that sample-window.
        example: 10 Hz samplerate will be binning 10k samples via mean() and
                 the timestamp for that new sample will be value[0] of the 10k long vector
    """

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        if self.delay and self.delay.total_seconds() < 0:
            raise ValueError("Delay can't be negative.")
        if self.duration and self.duration.total_seconds() < 0:
            raise ValueError("Duration can't be negative.")

        rates_allowed = (10, 100, 1_000, 100_000)
        if self.samplerate not in rates_allowed:
            raise ValueError(
                "Feature PowerTracing.samplerate only supports specific rates: %s", rates_allowed
            )
        return self


# NOTE: this was taken from pyserial (removes one dependency)
BAUDRATES = (
    50,
    75,
    110,
    134,
    150,
    200,
    300,
    600,
    1200,
    1800,
    2400,
    4800,
    9600,
    19200,
    38400,
    57600,
    115200,
    230400,
    460800,
    500000,
    576000,
    921600,
    1000000,
    1152000,
    1500000,
    2000000,
    2500000,
    3000000,
    3500000,
    4000000,
)

PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE = "N", "E", "O", "M", "S"
PARITIES = (PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE)

STOPBITS_ONE, STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO = (1, 1.5, 2)
STOPBITS = (STOPBITS_ONE, STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO)


class UartLogging(ShpModel, title="Config for UART Logging"):
    """Configuration for recording UART-Output of the Target Nodes.

    Note that the Communication has to be on a specific port that
    reaches the hardware-module of the SBC.
    """

    baudrate: Annotated[int, Field(ge=2_400, le=460_800)] = 115_200
    # ⤷ TODO: find maximum that the system can handle
    bytesize: Annotated[int, Field(ge=5, le=8)] = 8
    stopbits: Annotated[float, Field(ge=1, le=2)] = 1
    parity: str = PARITY_NONE

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        if self.baudrate not in BAUDRATES:
            msg = f"Error in config -> baud-rate must be one of: {BAUDRATES}"
            raise ValueError(msg)
        if self.stopbits not in STOPBITS:
            msg = f"Error in config -> stop-bits must be one of: {STOPBITS}"
            raise ValueError(msg)
        if self.parity not in PARITIES:
            msg = f"Error in config -> parity must be one of: {PARITIES}"
            raise ValueError(msg)
        return self


GpioInt = Annotated[int, Interval(ge=0, le=17)]
GpioList = Annotated[list[GpioInt], Field(min_length=1, max_length=18)]
all_gpio = list(range(18))


class GpioTracing(ShpModel, title="Config for GPIO-Tracing"):
    """Configuration for recording the GPIO-Output of the Target Nodes.

    TODO: postprocessing not implemented ATM
    """

    gpios: GpioList = all_gpio
    """List of GPIO to record.

    This feature allows to remove unwanted pins from recording,
    i.e. for chatty pins with separate UART Logging enabled.
    Numbering is based on the Target-Port and its 16x GPIO and two PwrGood-Signals.
    See doc for nRF_FRAM_Target_v1.3+ to see mapping of target port.

    Example for skipping UART (pin 0 & 1):
    .gpios = range(2,18)

    Note:
    - Cape 2.4 (2023) and lower only has 9x GPIO + 1x PwrGood
    - Cape 2.5 (2025) has first 12 GPIO & both PwrGood
    - this will be mapped accordingly by the observer
    """

    # time
    delay: timedelta = zero_duration
    duration: timedelta | None = None  # till EOF

    # post-processing,
    uart_decode: bool = False
    """Automatic decoding from gpio-trace not implemented ATM."""
    uart_pin: GPIO = GPIO(name="GPIO8")
    uart_baudrate: Annotated[int, Field(ge=2_400, le=1_152_000)] = 115_200

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        if self.delay and self.delay.total_seconds() < 0:
            raise ValueError("Delay can't be negative.")
        if self.duration and self.duration.total_seconds() < 0:
            raise ValueError("Duration can't be negative.")
        if self.uart_decode:
            log.error(
                "Feature GpioTracing.uart_decode reserved for future use. "
                "Use UartLogging or manually decode serial with the provided waveform decoder."
            )
        return self

    @property
    def gpio_mask(self) -> int:
        # valid for cape v2.5
        mask = 0
        for gpio in set(self.gpios):
            mask |= 2**gpio
        return mask


class GpioLevel(str, Enum):
    """Options for setting the gpio-level or state."""

    low = "L"
    high = "H"
    toggle = "X"  # TODO: not the smartest decision for writing a converter


class GpioEvent(ShpModel, title="Config for a GPIO-Event"):
    """Configuration for a single GPIO-Event (Actuation)."""

    delay: PositiveFloat
    """ ⤷ from start_time

    - resolution 10 us (guaranteed, but finer steps are possible)
    """
    gpio: GPIO
    level: GpioLevel
    period: Annotated[float, Field(ge=10e-6)] = 1
    """ ⤷ time base of periodicity in s"""
    count: Annotated[int, Field(ge=1, le=4096)] = 1

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        if not self.gpio.user_controllable():
            msg = f"GPIO '{self.gpio.name}' in actuation-event not controllable by user"
            raise ValueError(msg)
        return self

    def get_events(self) -> np.ndarray:
        stop = self.delay + self.count * self.period
        return np.arange(self.delay, stop, self.period)


class GpioActuation(ShpModel, title="Config for GPIO-Actuation"):
    """Configuration for a GPIO-Actuation-Sequence."""

    # TODO: not implemented ATM - decide if pru control sys-gpio or
    # TODO: not implemented ATM - reverses pru-gpio (preferred if possible)

    events: Annotated[list[GpioEvent], Field(min_length=1, max_length=1024)]

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        msg = "not implemented ATM"
        raise ValueError(msg)

    def get_gpios(self) -> set:
        return {_ev.gpio for _ev in self.events}


class SystemLogging(ShpModel, title="Config for System-Logging"):
    """Configuration for recording Debug-Output of the Observers System-Services."""

    kernel: bool = True
    time_sync: bool = True
    sheep: bool = True
    sys_util: bool = True


# TODO: some more interaction would be good
#     - execute limited python-scripts
#     - send uart-frames



"""Script to determine proper parameters for the fixtures.

This is done by comparing data from datasheets or online sources
with the resulting behavior of the model.
"""

from datetime import timedelta

from virtual_storage_config import VirtualStorageConfig

from shepherd_core import log
from shepherd_core.vsource.virtual_storage_models_kibam import ModelKiBaMPlus
from shepherd_core.vsource.virtual_storage_simulator import StorageSimulator


def experiment_self_discharge_lead_acid() -> None:
    """Determine leakage resistor and plot behavior.

    Datasheet of LEOCH LP12-1.2AH expects 3 to 20 % discharge/month
    -> chosen 5 % for the fixtures
    """
    dt_s = 10
    SoC_init = 1.0
    SoC_final = 0.95
    duration = timedelta(weeks=4)
    R_leak1 = VirtualStorageConfig.lead_acid(q_mAh=1).calc_R_leak_battery(
        duration, SoC_0=SoC_init, SoC_final=SoC_final
    )
    R_leak2 = (
        VirtualStorageConfig.lead_acid(q_mAh=10).calc_R_leak_battery(
            duration, SoC_0=SoC_init, SoC_final=SoC_final
        )
        * 10
    )
    log.info("R_leak_lead \t= %.0f Ohm [@1mAh]", R_leak1)
    if round(R_leak1) != round(R_leak2):
        raise ValueError("Values should match")
    cfg1 = VirtualStorageConfig.lead_acid(q_mAh=1200)
    sim = StorageSimulator(
        models=[ModelKiBaMPlus(SoC_init=SoC_init, cfg=cfg1, dt_s=dt_s)],
        dt_s=dt_s,
    )

    def step(_t: float, _s: float, _v: float) -> float:
        return 0

    sim.run(fn=step, duration_s=duration.total_seconds())
    sim.plot(
        f"XP {cfg1.name}, self-discharge, "
        f"SoC {SoC_init:.3f} to {SoC_final:.3f} in {duration.total_seconds()} s"
    )


def experiment_self_discharge_lipo() -> None:
    """Determine leakage resistor and plot behavior.

    According to wikipedia a typical Lipo looses 5 % in 4 weeks.
    """
    dt_s = 10
    SoC_init = 1.0
    SoC_final = 0.95
    duration = timedelta(weeks=4)
    R_leak1 = VirtualStorageConfig.lipo(q_mAh=1).calc_R_leak_battery(
        duration, SoC_0=SoC_init, SoC_final=SoC_final
    )
    R_leak2 = (
        VirtualStorageConfig.lipo(q_mAh=10).calc_R_leak_battery(
            duration, SoC_0=SoC_init, SoC_final=SoC_final
        )
        * 10
    )
    log.info("R_leak_lipo \t= %.0f Ohm [@1mAh]", R_leak1)
    if round(R_leak1) != round(R_leak2):
        raise ValueError("Values should match")
    cfg1 = VirtualStorageConfig.lipo(q_mAh=860)
    sim = StorageSimulator(
        models=[ModelKiBaMPlus(SoC_init=SoC_init, cfg=cfg1, dt_s=dt_s)],
        dt_s=dt_s,
    )

    def step(_t: float, _s: float, _v: float) -> float:
        return 0

    sim.run(fn=step, duration_s=duration.total_seconds())
    sim.plot(
        f"XP {cfg1.name}, self-discharge, "
        f"SoC {SoC_init:.3f} to {SoC_final:.3f} in {duration.total_seconds()} s"
    )


def experiment_self_discharge_tantal_avx() -> None:
    """Discharge Experiment: measured 5V to 3V decrease in 100 s.

    https://github.com/orgua/bq_characteristics/tree/main/eval_kit_behavior_var1#capacitor
    Note: first approach should be correct, but the second matches the plot
    """
    dt_s = 0.1
    cfg0 = VirtualStorageConfig.capacitor(C_uF=100, V_rated=6.3)
    duration = timedelta(seconds=100)
    R_leak = cfg0.calc_R_leak_battery(duration, SoC_final=3.0 / 5.0)
    log.info("R_leak_tantal \t= %.0f Ohm [@1uF]", R_leak * 100)
    cfg1 = VirtualStorageConfig.capacitor(
        C_uF=100, V_rated=6.3, R_leak_Ohm=R_leak, name="AVX TAJB107M006RNJ"
    )
    SoC_init = cfg1.approximate_SoC(V_OC=5.0)
    sim = StorageSimulator(
        models=[ModelKiBaMPlus(SoC_init=SoC_init, cfg=cfg1, dt_s=dt_s)],
        dt_s=dt_s,
    )

    def step(_t: float, _s: float, _v: float) -> float:
        return 0

    sim.run(fn=step, duration_s=duration.total_seconds())
    sim.plot(f"XP Tantal AVX, self-discharge {duration.total_seconds()} s")


def experiment_self_discharge_mlcc_tayo() -> None:
    """Discharge Experiment: measured 5V to 1.8V decrease in 100 s.

    Taiyo Yuden, JMK316ABJ107ML
    https://github.com/orgua/bq_characteristics/tree/main/eval_kit_behavior_var1#capacitor
    """
    dt_s = 0.1
    duration = timedelta(seconds=100)
    cfg0 = VirtualStorageConfig.capacitor(C_uF=74, V_rated=6.3)
    R_leak = cfg0.calc_R_leak_battery(duration, SoC_final=1.8 / 5.0)
    log.info("R_leak_mlcc \t= %.0f Ohm [@1uF]", R_leak * 74)
    R_leak = cfg0.calc_R_leak_capacitor(duration, SoC_final=1.8 / 5.0)
    log.info("R_leak_mlcc \t= %.0f Ohm [@1uF]", R_leak * 74)
    cfg1 = VirtualStorageConfig.capacitor(
        C_uF=74, V_rated=6.3, R_leak_Ohm=R_leak, name="Taiyo JMK316ABJ107ML"
    )
    SoC_init = cfg1.approximate_SoC(V_OC=5.0)
    sim = StorageSimulator(
        models=[ModelKiBaMPlus(SoC_init=SoC_init, cfg=cfg1, dt_s=dt_s)],
        dt_s=dt_s,
    )

    def step(_t: float, _s: float, _v: float) -> float:
        return 0

    sim.run(fn=step, duration_s=duration.total_seconds())
    sim.plot(f"XP MLCC Tayo, self-discharge {duration.total_seconds()} s")


if __name__ == "__main__":
    experiment_self_discharge_lead_acid()
    experiment_self_discharge_lipo()
    experiment_self_discharge_tantal_avx()
    experiment_self_discharge_mlcc_tayo()



"""Configuration for the Observer in Emulation-Mode."""

import copy
from collections.abc import Set as AbstractSet
from datetime import datetime
from datetime import timedelta
from enum import Enum
from pathlib import Path
from pathlib import PurePosixPath
from typing import Annotated

from pydantic import Field
from pydantic import model_validator
from pydantic import validate_call
from typing_extensions import Self

from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.base.timezone import local_tz
from shepherd_core.data_models.content.virtual_source_config import VirtualSourceConfig
from shepherd_core.data_models.experiment.experiment import Experiment
from shepherd_core.data_models.experiment.observer_features import GpioActuation
from shepherd_core.data_models.experiment.observer_features import GpioTracing
from shepherd_core.data_models.experiment.observer_features import PowerTracing
from shepherd_core.data_models.experiment.observer_features import SystemLogging
from shepherd_core.data_models.experiment.observer_features import UartLogging
from shepherd_core.data_models.experiment.target_config import vsrc_neutral
from shepherd_core.data_models.testbed import Testbed
from shepherd_core.data_models.testbed.cape import TargetPort
from shepherd_core.logger import log

from .helper_paths import path_posix


class Compression(str, Enum):
    """Options for choosing a dataset-compression."""

    lzf = "lzf"  # not native hdf5
    gzip1 = gzip = default = 1  # higher compr & load
    null = None
    # NOTE: lzf & external file-compression (xz or zstd) work better than gzip
    #       -> even with additional compression


compressions_allowed: list = [None, "lzf", 1]
c_translate = {"lzf": "lzf", "1": 1, "None": None, None: None}


class EmulationTask(ShpModel):
    """Configuration for the Observer in Emulation-Mode."""

    # General config
    input_path: Path
    """ ⤷ hdf5 file containing harvesting data"""
    output_path: Path | None = None
    """ ⤷ dir- or file-path for storing the recorded data:

    - providing a directory -> file is named emu_timestamp.h5
    - for a complete path the filename is not changed except it exists and
      overwrite is disabled -> emu#num.h5
    TODO: should the output-path be mandatory?
    """
    force_overwrite: bool = False
    """ ⤷ Overwrite existing file"""
    output_compression: Compression | None = Compression.default
    """ ⤷ should be lzf, 1 (gzip level 1) or None (order of recommendation)"""
    time_start: datetime | None = None
    """ timestamp or unix epoch time, None = ASAP"""
    duration: timedelta | None = None
    """ ⤷ Duration of recording in seconds, None = till EOF"""

    # emulation-specific
    use_cal_default: bool = False
    """ ⤷ Use default calibration values, skip loading from EEPROM"""

    enable_io: bool = True
    # TODO: add direction of pins! also it seems error-prone when only setting _tracing
    """ ⤷ Switch the GPIO level converter to targets on/off

    pre-req for sampling gpio / uart,
    """
    io_port: TargetPort = TargetPort.A
    """ ⤷ Either Port A or B that gets connected to IO"""
    pwr_port: TargetPort = TargetPort.A
    """ ⤷ selected port will be current-monitored

    - main channel is nnected to virtual Source
    - the other port is aux
    """
    voltage_aux: Annotated[float, Field(ge=0, le=4.5)] | str = 0
    """ ⤷ aux_voltage options
    - 0-4.5 for specific const Voltage (0 V = disabled),
    - "buffer" will output intermediate voltage (storage cap of vsource),
    - "main" will mirror main target voltage
    """
    # sub-elements, could be partly moved to emulation
    virtual_source: VirtualSourceConfig = vsrc_neutral
    """ ⤷ Use the desired setting for the virtual source,

    provide parameters or name like BQ25570
    """

    power_tracing: PowerTracing | None = PowerTracing()
    gpio_tracing: GpioTracing | None = GpioTracing()
    uart_logging: UartLogging | None = UartLogging()
    gpio_actuation: GpioActuation | None = None
    sys_logging: SystemLogging | None = SystemLogging()

    verbose: Annotated[int, Field(ge=0, le=4)] = 2
    """ ⤷ 0=Errors, 1=Warnings, 2=Info, 3=Debug,

    TODO: just bool now, systemwide
    """

    @model_validator(mode="before")
    @classmethod
    def pre_correction(cls, values: dict) -> dict:
        # convert & add local timezone-data
        has_time = values.get("time_start") is not None
        if has_time and isinstance(values["time_start"], (int, float)):
            values["time_start"] = datetime.fromtimestamp(values["time_start"], tz=local_tz())
        if has_time and isinstance(values["time_start"], str):
            values["time_start"] = datetime.fromisoformat(values["time_start"])
        if has_time and values["time_start"].tzinfo is None:
            values["time_start"] = values["time_start"].astimezone()
        return values

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        time_now = datetime.now().astimezone() - timedelta(hours=24)
        if self.time_start is not None and self.time_start < time_now:
            msg = (
                "Start-Time for Emulation can't be in the past "
                f"('{self.time_start}' vs '{time_now}'."
            )
            log.error(msg)  # do not raise anymore - server & clients do not have to match
        if self.duration and self.duration.total_seconds() < 0:
            raise ValueError("Task-Duration can't be negative.")
        if isinstance(self.voltage_aux, str) and self.voltage_aux not in {
            "main",
            "buffer",
        }:
            raise ValueError("Voltage Aux must be in float (0 - 4.5) or string 'main' / 'mid'.")
        if self.gpio_actuation is not None:
            raise ValueError("GPIO Actuation not yet implemented!")

        io_requested = any(
            _io is not None for _io in (self.gpio_actuation, self.gpio_tracing, self.uart_logging)
        )
        if self.enable_io and not io_requested:
            log.warning("Target IO enabled, but no feature requested IO")
        if not self.enable_io and io_requested:
            log.warning("Target IO not enabled, but a feature requested IO")
        return self

    @classmethod
    @validate_call
    def from_xp(cls, xp: Experiment, tb: Testbed, tgt_id: IdInt, root_path: Path) -> Self:
        obs = tb.get_observer(tgt_id)
        tgt_cfg = xp.get_target_config(tgt_id)
        io_requested = any(
            _io is not None
            for _io in (tgt_cfg.gpio_actuation, tgt_cfg.gpio_tracing, tgt_cfg.uart_logging)
        )

        return cls(
            input_path=path_posix(tgt_cfg.energy_env.data_path),
            output_path=path_posix(root_path / f"emu_{obs.name}.h5"),
            time_start=copy.copy(xp.time_start),
            duration=xp.duration,
            enable_io=io_requested,
            io_port=obs.get_target_port(tgt_id),
            pwr_port=obs.get_target_port(tgt_id),
            virtual_source=tgt_cfg.virtual_source,
            power_tracing=tgt_cfg.power_tracing,
            gpio_tracing=tgt_cfg.gpio_tracing,
            uart_logging=tgt_cfg.uart_logging,
            gpio_actuation=tgt_cfg.gpio_actuation,
            sys_logging=xp.sys_logging,
        )

    def is_contained(self, paths: AbstractSet[PurePosixPath]) -> bool:
        """Limit paths to allowed directories.

        TODO: could be added to validator (with a switch)
        """
        all_ok = any(self.input_path.is_relative_to(path) for path in paths)
        all_ok &= any(self.output_path.is_relative_to(path) for path in paths)
        return all_ok


# TODO: herdConfig
#  - store if path is remote (read & write)
#   -> so files need to be fetched or have a local path



"""AbstractBase-Class & Client-Class to access the file based fixtures.

Fixtures == OffLineDemoInstances
offline:	core  - fixtClient
webDev:	    core  - webClient  <->	webSrv - fixtClient
webUser:	core  - webClient  <->  webSrv - DbClient
webInfra:   core  - webClient+ <->  webSrv - DbClient

Users, Sheep and ServerApps should have access to the same DB via WebClient

Note: ABC and FixClient can't be in separate files when tb_client should
      default to FixClient (circular import)

TODO: Comfort functions missing
    - fixtures to DB, and vice versa
"""

from abc import ABC
from abc import abstractmethod
from typing import Any

from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.base.wrapper import Wrapper

from .fixtures import Fixtures


class AbcClient(ABC):
    """AbstractBase-Class to access a testbed instance."""

    def __init__(self) -> None:
        global tb_client  # noqa: PLW0603
        tb_client = self

    @abstractmethod
    def insert(self, data: ShpModel) -> bool:
        """Insert (and probably replace) entry.

        TODO: fixtures get replaced, but is that wanted for web?
        """

    @abstractmethod
    def query_ids(self, model_type: str) -> list[int]:
        pass

    @abstractmethod
    def query_names(self, model_type: str) -> list[str]:
        pass

    @abstractmethod
    def query_item(self, model_type: str, uid: int | None = None, name: str | None = None) -> dict:
        pass

    @abstractmethod
    def try_inheritance(
        self, model_type: str, values: dict[str, Any]
    ) -> tuple[dict[str, Any], list[str]]:
        # TODO: maybe internal? yes
        pass

    def try_completing_model(
        self, model_type: str, values: dict[str, Any]
    ) -> tuple[dict[str, Any], list[str]]:
        """Init by name/id, for none existing instances raise Exception.

        This is the main entry-point for querying a model (used be the core-lib).
        """
        if len(values) == 1 and next(iter(values.keys())) in {"id", "name"}:
            try:
                values = self.query_item(model_type, name=values.get("name"), uid=values.get("id"))
            except ValueError as err:
                msg = f"Query {model_type} by name / ID failed - {values} is unknown!"
                raise ValueError(msg) from err
        return self.try_inheritance(model_type, values)

    @abstractmethod
    def fill_in_user_data(self, values: dict[str, Any]) -> dict[str, Any]:
        # TODO: is it really helpful and needed?
        pass


class FixturesClient(AbcClient):
    """Client-Class to access the file based fixtures."""

    def __init__(self) -> None:
        super().__init__()
        self._fixtures: Fixtures = Fixtures()

    def insert(self, data: ShpModel) -> bool:
        wrap = Wrapper(
            datatype=type(data).__name__,
            parameters=data.model_dump(),
        )
        self._fixtures.insert_model(wrap)
        return True

    def query_ids(self, model_type: str) -> list[int]:
        return list(self._fixtures[model_type].elements_by_id.keys())

    def query_names(self, model_type: str) -> list[str]:
        return list(self._fixtures[model_type].elements_by_name.keys())

    def query_item(self, model_type: str, uid: int | None = None, name: str | None = None) -> dict:
        if uid is not None:
            return self._fixtures[model_type].query_id(uid)
        if name is not None:
            return self._fixtures[model_type].query_name(name)
        raise ValueError("Query needs either uid or name of object")

    def try_inheritance(
        self, model_type: str, values: dict[str, Any]
    ) -> tuple[dict[str, Any], list[str]]:
        return self._fixtures[model_type].inheritance(values)

    def fill_in_user_data(self, values: dict[str, Any]) -> dict[str, Any]:
        """Add fake user-data when offline-client is used.

        Hotfix until WebClient is working.
        """
        if values.get("owner") is None:
            values["owner"] = "unknown"
        if values.get("group") is None:
            values["group"] = "unknown"
        return values


tb_client: AbcClient = FixturesClient()
